from __future__ import print_function
import numpy as np
import scipy.linalg as lg
from . import wannier
from . import multicell
from . import angular
from scipy.sparse import csc_matrix, bmat

debug = True

if debug:
  fdebug = open(".debug.summetrize","w")

# routines to get a symmetric Hamiltonian from Wannier

def read_orbitals(input_file="orbitals.wan"):
  """Get the indexes and orbitals"""
  lines = open(input_file,"r").readlines()
  inds = [] # index
  atoms = [] # name
  orbs = [] # orbital
  rs = [] # coordinates in unit cell vectors
  for l in lines: # loop over lines
    l = l.split() # separte
    if "#" in l[0]: continue # comment, next iteration
    i = int(l[0])-1 # index
    ss = l[1] # atom and orbital
    ss = ss.split(",") # separate by the comma
    iatom = ss[0] # name of the atom
    iorb = ss[1] # name of the orbital
    inds.append(i) # store
    atoms.append(iatom) # store
    orbs.append(iorb) # store
    r = np.array([float(l[5]),float(l[6]),float(l[7])])
    rs.append(r) # store position
  return (inds,atoms,orbs,rs) # return the three things


def index_dictionary(inds,atoms,orbs):
  """Get a dictionary with the indexes of atom and orbital"""
  idict = dict() # create dictionary
  for (i,a,o) in zip(inds,atoms,orbs): # loop over indexes
    idict[(a,o)] = i # store index for atom and orbital
  return idict # return the dictionary


def orbital_dictionary(inds,atoms,orbs):
  """Get a dictionary with the indexes of atom and orbital"""
  odict = dict() # create dictionary
  for a in atoms: odict[a] = [] # empty list, initialize
  for (a,o) in zip(atoms,orbs): # loop over orbitals
    odict[a] += [o] # add this orbital
  return odict # return the dictionary

def get_atom_names(atoms):
  """Get names of the atoms"""
  anames = []
  for a in atoms:
    if  not a in anames: anames.append(a)
  return anames # return names



def position_dictionary(atoms,rs):
  """Get names of the atoms"""
  rdict = dict() # dictionary
  for (a,r) in zip(atoms,rs):
    rdict[a] = r # store position
  return rdict # return dictionary



def read_unit_cell(input_file="geometry.wan"):
  """Read unit cell vectors"""
  m = np.genfromtxt(input_file)
  return (m[0],m[1],m[2]) # return vectors



def extract_matrix(a1,a2,odict,idict,matrix):
  """Extract the hopping between two atoms"""
  orbs1 = odict[a1] # orbitals of the first atom
  orbs2 = odict[a2] # orbitals of the second atom
  n1 = len(orbs1) # size of the basis in the first atom
  n2 = len(orbs2) # size of the basis in the second atom
  zeros = np.matrix(np.zeros((n1,n2),dtype=np.complex)) # zero matrix  
#  print(a1,a2,orbs1)
  for ii in range(len(orbs1)): # loop over orbitals in the first atom
    for jj in range(len(orbs2)): # loop over orbitals in the second atom
      i = idict[(a1,orbs1[ii])] # index of this orbital 
      j = idict[(a2,orbs2[jj])] # index of this orbital 
      zeros[ii,jj] = matrix[i,j] # store this matrix
#      print(a1,a2,i,j,orbs1[ii],orbs2[jj])
#  raise
  return zeros # return matrix





def rotation_operator(orbs,v=np.array([1.,0.,0.])):
  """Generate a rotation matrix along the z axis, depending on the
  orbitals used. Rotates vector v to vector 1,0,0"""
  n = len(orbs) # number of orbitals
  zeros = np.matrix(np.zeros((n,n),dtype=np.complex)) # zero matrix  
  (lx,ly,lz) = angular.angular_momentum(orbs) # angular momentum matrices
  v = np.array(v)
  vv = v.dot(v) # modulus
  if vv> 0.000001:
    v = v/(np.sqrt(v.dot(v))) # normalize the vector
    angle = np.arctan2(v[1],v[0]) # get the angle
  else:
    angle = 0.0 # zero angle, to return identity
#  raise
  R = np.matrix(lg.expm(-1j*lz*angle)) # get the rotation matrix
  if np.sum(np.abs(lz-lz.H))>0.0001: raise # check that lz is Hermitian
  return R # return rotation operator
  
  


def reorder_basis(anames,odict,idict):
  """Matrix that transforms the new reordered basis into the original one"""
  n = sum([len(odict[a]) for a in anames]) # length of the basis
  zeros = np.matrix(np.zeros((n,n),dtype=np.complex)) # zero matrix  
  i = 0 # initialize counter
  for a in anames: # loop over atoms
    for o in odict[a]: # loop over orbitals
      j = idict[(a,o)] # get index in the Wannier file
      zeros[i,j] = 1.
      i += 1 # increase counter
  return zeros


def local_symmetrizer(anames,ons,odict,sym_file="symmetry.wan"):
  """Return the matrices that symmetrizes the local crystal field,
  mposing a certain symmetry"""
  stypedict = dict() # dictionary for the symmetries
  sdict = dict() # dictionary for the symmetry matrices
  lines = open(sym_file).readlines() # read the lines
  for l in lines:
    l = l.split()
    stypedict[l[0]] = l[1] # save symmetry for this atom
  for (a,m) in zip(anames,ons): # now do a loop over all the different atoms
    sname = stypedict[a] # symmetry for this atom
    if sname == "C3": nr = 3# C3 symmetry
    else: raise # not implemented
    step = 2.*np.pi/nr # 2 times pi over number of rotations
    # generate vectors
    vs = [np.array([np.cos(step*i),np.sin(step*i)]) for i in range(nr)]
    # generate rotation matrices
    Rs = [rotation_operator(odict[a],v=v) for v in vs]
    # perform all the rotations in the onsite matrix
    hrs = [R.H*m*R for R in Rs]
    hr,error = average_matrices(hrs) # perform the average
    print("Broken symmetry in",a,"is",error)
    sdict[a] = hr # store rotation matrix
  return sdict # return the dictionary
  



def symmetrize_hamiltonian(orb_file="orbitals.wan",ham_file="hamiltonian.wan",
                        cell_file="geometry.wan",nlook=[4,4,1],
                        sym_file="symmetry.wan"):
  """Write in a file the new Hamiltonian, symmetrized"""
  (inds,atoms,orbs,rs) = read_orbitals(input_file=orb_file) # read the orbitals
  anames = get_atom_names(atoms) # different atom names
  # function to get hoppings to arbitrary cells
  get_hopping = multicell.read_from_file(input_file=ham_file)
  # get the dictionaries
  idict = index_dictionary(inds,atoms,orbs)
  odict = orbital_dictionary(inds,atoms,orbs)
  rdict = position_dictionary(atoms,rs)
  # matrix that will reorder the orbitals
  T = reorder_basis(anames,odict,idict)
  # read unit cell vectors
  (a1,a2,a3) = read_unit_cell()
  # generate the symmetrizer matrices
  h0 = get_hopping(0,0,0) # onsite matrix
  ons = [extract_matrix(a,a,odict,idict,h0) for a in anames] # in each atom
  sdict = local_symmetrizer(anames,ons,odict,sym_file=sym_file) # matrices
  # now look into neighbors to get matrices 
  stored_rs = [] # distances between atoms stored
  stored_atoms = [] # distances between atoms stored
  stored_matrix = [] # distances between atoms stored
  stored_vectors = [] # vectors between atoms
  for ia in anames: # loop over atoms
    for ja in anames: # loop over atoms
      ri = rdict[ia] # position of iesim atom
      ri = ri[0]*a1 + ri[1]*a2 + ri[2]*a3 # position in real coordinates
      for n1 in range(-nlook[0],nlook[0]+1):
        for n2 in range(-nlook[1],nlook[1]+1):
          for n3 in range(-nlook[2],nlook[2]+1):
            rj = rdict[ja] # position of iesim atom
            # position in real coordinates
            rj = (rj[0]+n1)*a1 + (rj[1]+n2)*a2 + (rj[2]+n3)*a3 
            dr = rj-ri # vector between this two atoms 
            drr = np.sqrt(dr.dot(dr)) # distance between atoms
            m = get_hopping(n1,n2,n3) # hopping in this direction
            tij = extract_matrix(ia,ja,odict,idict,m) # hopping between atoms
            # special case of onsite matrices
            if ia==ja and n1==n2==n3==0:
              tij = sdict[ia] # get the symmetrized onsite matrix
            Ri = rotation_operator(odict[ia],dr) # rotate basis i
            Rj = rotation_operator(odict[ja],dr) # rotate basis j
            Rtij = Ri.H*tij*Rj # hopping matrix in the new frame
            # now store the different things needed
            stored_rs.append(drr) # store the distance
            stored_vectors.append(dr) # store the vector
            stored_atoms.append((ia,ja)) # store the atom pairs
            stored_matrix.append(Rtij) # store the rotated matrix
  # get the function that generated symmetric hoppings
  hop_gen = symmetric_hopping_generator(stored_rs,stored_vectors,
                                          stored_atoms,stored_matrix)
  # now create hoppings of the symmetric Hamiltonian
  rmax = np.sqrt(a1.dot(a1))*(nlook[0]+1)/2. # maximum distance
  print("Cutoff distance is",rmax)
  ns_list = [] # list with ns
  hops_list = [] # list with ns
  for n1 in range(-nlook[0],nlook[0]+1):
    for n2 in range(-nlook[1],nlook[1]+1):
      for n3 in range(-nlook[2],nlook[2]+1):
        hop = [[None for i in range(len(anames))] for j in range(len(anames))]
        for i in range(len(anames)): # loop over atoms
          for j in range(len(anames)): # loop over atoms
            ia = anames[i] # name of the atom
            ja = anames[j] # name of the atom
            ri = rdict[ia] # position of iesim atom
            rj = rdict[ja] # position of jesim atom
            ri = ri[0]*a1 + ri[1]*a2 + ri[2]*a3 # position in real coordinates
            rj = (rj[0]+n1)*a1 + (rj[1]+n2)*a2 + (rj[2]+n3)*a3 
            dr = rj-ri # vector between this two atoms 
            drr = np.sqrt(dr.dot(dr)) # distance between atoms
            m = hop_gen(ia,ja,dr)# get the symmetric hopping
            Ri = rotation_operator(odict[ia],dr) # rotate basis i
            Rj = rotation_operator(odict[ja],dr) # rotate basis j
            if m is not None:
              m = Ri*m*Rj.H # hopping matrix in the new frame, opposite rot
              if drr>rmax: m *= 0. # reached maximum distance
              hop[i][j] = csc_matrix(m) # store the matrix 
        hop = bmat(hop).todense() # create dense matrix
        hop = T.H *hop *T # convert to the original order
        hops_list.append(hop) # store hopping
        ns_list.append([n1,n2,n3]) # store indexes
  multicell.save_multicell(ns_list,hops_list)
      

def symmetric_hopping_generator(rs,drs,atoms,matrix):
  """Returns a function that generates hoppings with the symmetry of the
  lattice, inputs are names of the atoms and distance between them"""
  ijdict = dict() # dictionary for hopping between atoms ij
  for ij in atoms: ijdict[ij] = [] # initialize as empty list
  # now do a loop over all inputs, storing the matrices
  for (r,dr,ij,m) in zip(rs,drs,atoms,matrix): # loop over inputs
    ijdict[ij].append((r,dr,m)) # store distance and matrix
  ijs = [] # empty list
  for ij in atoms: # loop over atoms
    if ij not in ijs: ijs.append(ij) # store the different pairs
  tdict = dict() # dictionary for the hoppings
  for ij in ijs: # loop over different pairs of atoms
    rm = ijdict[ij] # get the distances and matrices for this pair
    difr = [r for (r,dr,m) in rm] # list with different distances
    difdr = [dr for (r,dr,m) in rm] # list with different distances
    # function to return the neighbor index
    (get_index,nmax) = get_neighbor_order(difdr,sym="C3") # get the index 
    nmat = [[] for i in range(nmax)] # initilize list
    for (r,dr,m) in rm: # loop over matrices of this atom
      nmat[get_index(dr)].append(m) # store this matrix
    mdms = [] # empty list
    print("Found ",len(nmat[1]),"matrices for ",ij)
    for i1 in range(nmax): # loop over distances
# calculate all the averages
      mdms.append(average_matrices(nmat[i1])) 
    dms = [dm[1] for dm in mdms] # store errors
    ms = [im[0] for im in mdms] # store matrices
    if np.max(dms)>0.0001: print(ij,"is not symmetric",np.max(dms))
    else: print(ij,"is symmetric")
    tdict[ij] = (get_index,ms) # store in the dictionary
  def get_symmetric_hopping(i,j,r):
    """Function that return a hopping that respects the crystal symmetries"""
    (gi,ms) = tdict[(i,j)] # get the right set
    try: return ms[gi(r)] # return the right matrix
    except: return None # return None
  return get_symmetric_hopping # return the function
    
    


def get_neighbor_order(drs,sym="C3",prec=1):
  """Returns a function that given an input distance, says the order
  of enighbor it is"""
  delta = 10**(-prec) # precision
  if sym=="C3":  cn = 3 # group symmetry used
  elif sym=="C6":  cn = 6 # group symmetry used
  elif sym=="C4":  cn = 4 # group symmetry used
  alpha = 2.*np.pi/cn # C3 rotation
  sa,ca = np.sin(alpha), np.cos(alpha) # sin and cosine
  Rot = np.matrix([[ca,sa,0.],[-sa,ca,0.],[0.,0.,1.]]) # rotation matrix
  Rs = [Rot**i for i in range(cn)] # rotation matrices
  def same_vectors(dr,rs):
    """Check if two vectors are related by a symmetry operation"""
    norm1 = np.sqrt(dr.dot(dr)) # first norm
    ur1 = dr/norm1 # unit vector
    norm2 = np.sqrt(rs.dot(rs)) # secon norm
    if norm1==0. and norm2==0.: return True # special case of 0 vector
    if not np.abs(norm1-norm2)<delta: return False # different distance
    for R in Rs: # loop over operations
      rs2 = R*(np.matrix([rs]).T) # rotate    
      rs2 = np.array(rs2).reshape(3) # convert to 1d array
      diff = dr - rs2 # vector difference
      if diff.dot(diff)<delta:
#        print("Same",dr-rs)
        return True # this vector in not independent    
    return False # the vectors are different
  out = [] # irreducible vectors
  dis = [] # distances
  for dr in drs: # loop over distances
    if not sum([same_vectors(dr,rs) for rs in out]): # vector not present 
      out.append(dr) # store this vector
      dis.append(dr.dot(dr)) # store this vector
  out = [o.tolist() for o in out] # convert to list to avoid bug
  out = [o for (d,o) in sorted(zip(dis,out))]
  out = [np.array(o) for o in out] # convert back to array
#'  print(sorted(dis))
#  out = zip(out,dis).sort(key=lambda (x,y): b.index(x))
  def get_group(r):
    """Return the group of neighbor"""
    for i in range(len(out)):
      if same_vectors(r,out[i]): return i # group of the vector    
    print("Not found",r,out)
    raise # raise error if not found
  return (get_group,len(out)) # return the function













def average_matrices(ms):
  """Calculate the average of certain matrices"""
  if len(ms)==0: return (None,0.)
  mo = 0.0*ms[0] # initialize
  for m in ms: mo += m # add matrix
  mo = mo/len(ms) # average
  dm = sum([np.sum(np.abs(m-mo)) for m in ms])/len(ms) # error
#  dm /= np.sum(np.abs(ms)) # normalize to the dimension
  return mo,dm







import numpy as np



def write_xyz(gin,output_file = "crystal.xyz",units=0.529,nsuper=1):
  """Writes the geometry associatted with a hamiltonian in a file"""
  g = gin.copy() # copy geometry
#  if g.atoms_have_names: g = remove_duplicated(g)
#  else: g = gin.copy()
#  g = gin.copy()
  if g.dimensionality>0: # create supercell
    if nsuper>1:
      g = g.supercell(nsuper)
  x = g.x*units  # x posiions
  y = g.y*units  # y posiions
  z = g.z*units  # z posiions
  fg = open(output_file,"w")
  # get hte names of the atoms
  if g.atoms_have_names:
    names = g.atoms_names
  else:
    names = ["C" for ix in x ] # create names
  # check that there are as many names as positions
  if len(names)!=len(x): raise
  fg.write(str(len(x))+"\nGenerated with Python\n") # number of atoms
  for (n,ix,iy,iz) in zip(names,x,y,z):
    fg.write(n+"   "+str(ix)+ "     "+str(iy)+"   "+str(iz)+"  \n")
  fg.close()






def write_lattice(g,output_file = "LATTICE.OUT"):
  """Writes the lattice in a separte file"""
  open("DIMENSIONALITY.OUT","w").write(str(g.dimensionality))
  if g.dimensionality==0: return
  fg = open(output_file,"w")
  if g.dimensionality==1:
    fg.write(str(g.a1[0])+"   "+str(g.a1[1])+"  "+str(g.a1[2])+"\n")
  elif g.dimensionality==2:
    fg.write(str(g.a1[0])+"   "+str(g.a1[1])+"  "+str(g.a1[2])+"\n")
    fg.write(str(g.a2[0])+"   "+str(g.a2[1])+"  "+str(g.a2[2])+"\n")
  elif g.dimensionality==3:
    fg.write(str(g.a1[0])+"   "+str(g.a1[1])+"  "+str(g.a1[2])+"\n")
    fg.write(str(g.a2[0])+"   "+str(g.a2[1])+"  "+str(g.a2[2])+"\n")
    fg.write(str(g.a3[0])+"   "+str(g.a3[1])+"  "+str(g.a3[2])+"\n")
  else: raise
  fg.close()



def write_sublattice(g,output_file = "SUBLATTICE.OUT"):
  """Writes the geometry associatted with a hamiltonian in a file"""
  if not g.has_sublattice: return
  fg = open(output_file,"w")
  for i in g.sublattice:
    fg.write(str(i)+"  \n")
  fg.close()



def write_positions(g,output_file = "POSITIONS.OUT",nrep=None):
  """Writes the geometry associatted with a hamiltonian in a file"""
  if nrep is not None: g = g.get_supercell(nrep)
  x = g.r[:,0]  # x positions
  y = g.r[:,1]  # y positions
  z = g.r[:,2]  # z positions
  fg = open(output_file,"w")
  fg.write(" # x    y     z   (without spin degree)\n")
  for (ix,iy,iz) in zip(x,y,z):
    fg.write(str(ix)+ "     "+str(iy)+"   "+str(iz)+"  \n")
  fg.close()





def write_vasp(g0,s=1.42):
    """Turn a geometry into vasp geometry"""
    g = g0.copy() # copy geometry
    if g.dimensionality==3: pass
    elif g.dimensionality==2:
        g.r[:,2] -= np.min(g.r[:,2])
        z = np.max(g.r[:,2]) - np.min(g.r[:,2])
        a3 = np.array([0.,0.,z+9.0])
        g.a3 = a3 # set the lattice vector
        g.dimensionality = 3
        g.get_fractional() # get fractional coordinates
    else: raise # not implemented
    f = open("vasp.vasp","w") # input file
    f.write("Structure\n 1.0\n")
    for i in range(3): f.write(str(s*g.a1[i])+"  ")
    f.write("\n")
    for i in range(3): f.write(str(s*g.a2[i])+"  ")
    f.write("\n")
    for i in range(3): f.write(str(s*g.a3[i])+"  ")
    # write the atoms
    if not g.atoms_have_names: # no name provided
      f.write("\n C\n "+str(len(g.r))+"\n Direct\n")
      # write all the atoms in fractional coordinates
      for ir in g.frac_r:
          for i in range(3): f.write(str(ir[i])+"  ")
          f.write("\n")
    else: # atoms have labels
        namedict = dict() # dictionary
        for (key,n) in zip(g.atoms_names,range(len(g.r))):
            if not key in namedict:
                namedict[key] = [n] # store
            else: namedict[key].append(n) # store
        f.write("\n") # next line
        for key in namedict: f.write(str(key)+"   ")
        f.write("\n") # next line
        for key in namedict: f.write(str(len(namedict[key]))+"   ")
        f.write("\n Direct \n") # next line
        for key in namedict: # loop over types
            ns = namedict[key] # list with atoms
            for ii in ns: # loop over atoms
              for i in range(3): f.write(str(g.frac_r[ii][i])+"  ")
              f.write("\n")
    f.close()




def write_function(self,fun,name="FUNCTION.OUT"):
    """Write a certain function"""
    f = open(name,"w")
    ir = 0
    for r in self.r: # loop over positions
      o = fun(r) # evaluate
      f.write(str(ir)+"  ")
      for io in o:  f.write(str(io)+"  ")
      f.write("\n")
      ir += 1
    f.close() # close file





import numpy as np

from .. import specialhopping

def get_hamiltonian(self,tij=None,fun=None,has_spin=True,ts=None,
                        is_sparse=False,spinful_generator=False,nc=2,
                        is_multicell=False,mgenerator=None,**kwargs):
    """ Create the hamiltonian for this geometry. By default, it assumes
    first neighbor hopping
        - has_spin = True, whether if to include spin degree of freedom
        - is_sparse = False, use sparse representation
        - ts = None, list of 1st,2nd,3rd neighbor hoppings
        - tij = None, function that return the spatially dependent hoppings
    """
    ### Perform some initial sanity checks
    if fun is not None: 
        print("fun will be deprecated, use instead tij")
        tij = fun # overwrite
    ## in case tij is an iterable with hoppings
    from ..checkclass import is_iterable
    if is_iterable(tij): # tij is an iterable
        ts = tij # overwrite
        tij = None
    #####
    if ts is not None: # if hoppings given, overwrite mgenerator
        mgenerator = specialhopping.neighbor_hopping_matrix(self,ts)
        is_multicell = True # overwrite
        nc = 2*len(ts) + 1 # overwrite
    if self.dimensionality==3: is_multicell=True
    from ..hamiltonians import Hamiltonian
    h = Hamiltonian(self)  # create the object
    h.is_sparse = is_sparse
    h.has_spin = has_spin
    h.is_multicell = is_multicell
    if is_multicell:  # workaround for multicell hamiltonians
      from ..multicell import parametric_hopping_hamiltonian
      if mgenerator is not None:
          from ..multicell import parametric_matrix # not implemented
          h = parametric_matrix(h,fm=mgenerator,cutoff=nc,**kwargs)
      else: h = parametric_hopping_hamiltonian(h,fc=tij,**kwargs) # add hopping
      return h
    if tij is None and mgenerator is None: # no function given
      h.first_neighbors()  # create first neighbor hopping
    else: # function or mgenerator given
      if h.dimensionality<3:
        from ..hamiltonians import generate_parametric_hopping
        h = generate_parametric_hopping(h,f=tij,
                  spinful_generator=spinful_generator,
                  mgenerator=mgenerator) # add hopping
      elif h.dimensionality==3:
        if mgenerator is not None: raise # not implemented
        from ..multicell import parametric_hopping_hamiltonian
        h = parametric_hopping_hamiltonian(h,fc=tij,**kwargs) # add hopping
    if not is_sparse: h.turn_dense() # dense Hamiltonian
    return h # return the object




import numpy as np
from ..superconductivity import get_eh_sector
from ..superconductivity import build_nambu_matrix
from ..multihopping import MultiHopping
from .. import algebra

def extract_anomalous_dict(dd):
    """Given a dictionary, extract the anomalous part"""
    out = dict()
    for key in dd:
        d = dd[key] # get this patrix
        m01 = get_eh_sector(d,i=0,j=1)
        m10 = get_eh_sector(d,i=1,j=0)
        m = build_nambu_matrix(m01*0.0,c12=m01,c21=m10) # build matrix
        out[key] = m
    return out # return dictionary


def extract_normal_dict(dd):
    """Given a dictionary, extract the anomalous part"""
    out = dict()
    for key in dd:
        d = dd[key] # get this patrix
        m00 = get_eh_sector(d,i=0,j=0)
        m = build_nambu_matrix(m00) # build matrix
        out[key] = m
    return out # return dictionary




def get_anomalous_hamiltonian(self):
    """Turn a Hamiltonian into a Nambu Hamiltonian"""
    self = self.copy() # copy Hamiltonian
    self.turn_nambu() # setup electron-hole if not present
    dd = self.get_multihopping().get_dict() # return the dictionary
    dd = extract_anomalous_dict(dd)
    self.set_multihopping(MultiHopping(dd))
    return self


def get_singlet_hamiltonian(self):
    self = self.copy()
    dd = self.get_multihopping().get_dict() # return the dictionary
    dd = extract_singlet_dict(dd)
    self.set_multihopping(MultiHopping(dd))
    return self


def get_triplet_hamiltonian(self):
    h = get_anomalous_hamiltonian(self)
    return h - get_singlet_hamiltonian(h)



def extract_pairing(m):
  """Extract the pairing from a matrix, assuming it has the Nambu form"""
  nr = m.shape[0]//4 # number of positions
  uu = np.array(np.zeros((nr,nr),dtype=np.complex)) # zero matrix
  dd = np.array(np.zeros((nr,nr),dtype=np.complex)) # zero matrix
  ud = np.array(np.zeros((nr,nr),dtype=np.complex)) # zero matrix
  for i in range(nr): # loop over positions
    for j in range(nr): # loop over positions
        ud[i,j] = m[4*i,4*j+2]
        dd[i,j] = m[4*i+1,4*j+2]
        uu[i,j] = m[4*i,4*j+3]
  return (uu,dd,ud) # return the three matrices



def extract_triplet_pairing(m):
  """Extract the pairing from a matrix, assuming it has the Nambu form"""
  m = algebra.todense(m) # dense matrix
  nr = m.shape[0]//4 # number of positions
  uu = np.array(np.zeros((nr,nr),dtype=np.complex)) # zero matrix
  dd = np.array(np.zeros((nr,nr),dtype=np.complex)) # zero matrix
  ud = np.array(np.zeros((nr,nr),dtype=np.complex)) # zero matrix
  for i in range(nr): # loop over positions
    for j in range(nr): # loop over positions
        ud[i,j] = (m[4*i,4*j+2] - np.conjugate(m[4*j+3,4*i+1]))/2.
        dd[i,j] = m[4*i+1,4*j+2]
        uu[i,j] = m[4*i,4*j+3]
  return (uu,dd,ud) # return the three matrices


def extract_singlet_dict(dd):
    """Given a dictionary, extract the anomalous singlet part"""
    out = dict()
    for key in dd:
        d = dd[key] # get this matrix
        nr = d.shape[0]//4 # number of positions
        m = np.zeros(d.shape,dtype=np.complex) # initialize
        m0 = dd[key]
        key2 = (-key[0],-key[1],-key[2]) 
        m1 = dd[key2]
        for i in range(nr): # loop over positions
            for j in range(nr): # loop over positions
                m[4*i,4*j+2] = (m0[4*i,4*j+2]+np.conjugate(m1[4*j+3,4*i+1]))/2.
                m[4*j+3,4*i+1] = (m0[4*j+3,4*i+1]+np.conjugate(m1[4*i,4*j+2]))/2.
                m[4*i+2,4*j] = (m0[4*i+2,4*j]+np.conjugate(m1[4*j+1,4*i+3]))/2.
                m[4*j+1,4*i+3] = (m0[4*j+1,4*i+3]+np.conjugate(m1[4*i+2,4*j]))/2.
        out[key] = m
    return out # return dictionary


def extract_triplet_dict(dd):
    dd = extract_anomalous_dict(dd) # anomalous part
    ds = extract_singlet_dict(dd) # singlet part
    from ..multihopping import MultiHopping
    return (MultiHopping(dd) - MultiHopping(ds)).get_dict()



def extract_singlet_pairing(m):
  """Extract the pairing from a matrix, assuming it has the Nambu form"""
  nr = m.shape[0]//4 # number of positions
  ud = np.array(np.zeros((nr,nr),dtype=np.complex)) # zero matrix
  for i in range(nr): # loop over positions
    for j in range(nr): # loop over positions
        ud[i,j] = (m[4*i,4*j+2] + np.conjugate(m[4*j+3,4*i+1]))/2.
  return ud







from __future__ import print_function
import numpy as np
from . import geometry
from . import sculpt


def get_geometry(name="square",n=5,nedges=None,rot=0.,clean=True,
                 geo=None,shift=[0.,0.,0.]):
  """ Create a 0d island"""
  lattice_name = name
  if lattice_name=="honeycomb":
    geometry_builder = geometry.honeycomb_lattice
    if nedges==None: nedges = 6
  elif lattice_name=="square":
    geometry_builder = geometry.square_lattice
    if nedges==None: nedges = 4
  elif lattice_name=="kagome":
    geometry_builder = geometry.kagome_lattice
    if nedges==None: nedges = 3
  elif lattice_name=="lieb":
    geometry_builder = geometry.lieb_lattice
    if nedges==None: nedges = 4
  elif lattice_name=="triangular":
    geometry_builder = geometry.triangular_lattice
    if nedges==None: nedges = 3
  else: raise
  # first create a raw unit cell
  if geo is not None: 
    print("Geometry generator taken from input")
    g = geo # builder is input geometry
  else:
    g = geometry_builder()  # build a 2d unit cell
  nf = float(n)   # get the desired size, in float
  g = g.supercell(int(7*n))   # create supercell
  g.set_finite() # set as finite system
  g.center() # center the geometry
  # now scuplt the geometry
  g = sculpt.rotate(g,rot) # initial rotation
  # now shift the lattice
  g.x += shift[0]
  g.y += shift[1]
  g.z += shift[2]
  g.xyz2r()
  def f(r): return r[0]>-nf*(np.cos(np.pi/3)+1.)  # function to use as cut
  for i in range(nedges): # loop over rotations, 60 degrees
    g = sculpt.intersec(g,f) # retain certain atoms
    g = sculpt.rotate(g,2.*np.pi/nedges) # rotate 60 degrees
  if clean: # if it is cleaned
    g = sculpt.remove_unibonded(g,iterative=True)  # remove single bonded atoms
  g.center() # center the geometry
  return g






from __future__ import print_function
import scipy.linalg as lg
import scipy.sparse.linalg as slg
from scipy.sparse import csc_matrix,eye
import numpy as np

def correlator0d(m,energies=np.linspace(-10.,10.,400),i=0,j=0,delta=0.07):
  """Calculate a certain correlator"""
  iden = np.identity(m.shape[0],dtype=np.complex)
  zs = np.zeros(energies.shape[0],dtype=np.complex)
  for (ie,e) in zip(range(len(energies)),energies):
    m0 = ((e+1j*delta)*iden - m).I # inverse 
    zs[ie] = m0[i,j]
  np.savetxt("CORRELATOR.OUT",np.matrix([energies,zs.real,-zs.imag]).T)
  print("Saved correlator in CORRELATOR.OUT")
  return (energies,zs.real,-zs.imag)



def gs_correlator(m,i=0,j=0):
    """Compute the integrated correlator using wavefunctions"""
    (es,vs) = lg.eigh(m) # diagonalize
    vs = vs.transpose() # transpose matrix
    c = 0.0
    for (e,v) in zip(es,vs):
        if e<0.0:
            c += v[i]*np.conjugate(v[j])
    return c.real


def dm_ij_energy(m,ne=500,scale=10.,i=0,j=0,delta=0.07):
  """Calculate a certain correlator"""
  energies = np.linspace(-scale,scale,ne)
  iden = np.identity(m.shape[0],dtype=np.complex)
  zs0 = np.zeros(energies.shape[0],dtype=np.complex)
  zs1 = np.zeros(energies.shape[0],dtype=np.complex)
  for (ie,e) in zip(range(len(energies)),energies):
    m0 = ((e+1j*delta)*iden - m).I # inverse 
    m1 = ((e-1j*delta)*iden - m).I # inverse 
    zs0[ie] = m0[i,j]
    zs1[ie] = m1[i,j]
  z = (zs1 - zs0)/2. # both contributions
  return (energies,-z*1j)






def gij(m,i=0,delta=0.01,e=0.0):
  """Calculate a single row of the Green function"""
  v0 = np.zeros(m.shape[0])
  v0[i] = 1.
  iden = eye(v0.shape[0]) # identity matrix
  g = iden*(e+1j*delta) - csc_matrix(m) # matrix to invert
#  print(type(g)) ; exit()
  (b,info) = slg.lgmres(g,v0) # solve the equation  
  go = (b*np.conjugate(b)).real
  return go






import os
import numpy as np
from ..scftypes import scfclass
from ..scftypes import get_fermi_energy
from ..scftypes import get_occupied_states
from scipy.sparse import csc_matrix
import time
from ..scftypes import directional_mean_field
from .. import limits
from .. import inout


try:
  from .. import correlatorsf90
except: pass



mf_file = "MF.pkl" # mean field file

def hubbardscf(h,g=1.0,nkp = 100,filling=0.5,mag=None,mix=0.9,
                  maxerror=1e-05,silent=False,mf=None,
                  T=None,collinear=False,fermi_shift=0.0,
                  maxite=1000,save=False,nk=None,U=None):
  """ Solve a selfconsistent Hubbard mean field"""
  # alternative input variables
  if nk is not None: nkp = nk
  if U is not None: g = U
  #############################
  mix = 1. - mix
  U = g # redefine
  if not h.has_spin: raise
  os.system("rm -f STOP") # remove stop file
  from scipy.linalg import eigh
  nat = h.intra.shape[0]//2 # number of atoms
  htmp = h.copy()  # copy hamiltonian
  htmp.turn_dense() # turn into a dense Hamiltonian
  # generate the necessary list of correlators
  if mf is None: # generate initial mean field
    try:  
        old_mf = inout.load(mf_file) # load the file
        print("Mean field read from file")
    except: # generate mean field
      if mag is None: mag = np.random.random((nat,3))
# get mean field matrix
      old_mf = selective_U_matrix(U,directional_mean_field(mag))
  else: old_mf = mf # use guess
  # get the pairs for the correlators
  ndim = h.intra.shape[0] # dimension
  totkp = nkp**(h.dimensionality) # total number of kpoints
  file_etot = open("SCF_ENERGY.OUT","w")
  file_error = open("SCF_ERROR.OUT","w")
  ite = 0 # iteration counter
  scf = scfclass(h) # create scf class
  while True: # infinite loop
    ite += 1 # increase counter
    htmp.intra = h.intra + old_mf # add mean field 
    t1 = time.time()
# get eigenvectors
    eigvals,eigvecs,kvectors = htmp.get_eigenvectors(nk=nkp,kpoints=True)
    eigvecs = np.conjugate(eigvecs)
# fermi energy
    t2 = time.time()
    fermi = get_fermi_energy(eigvals,filling,fermi_shift=fermi_shift)
# occupied states
    eoccs,voccs,koccs = get_occupied_states(eigvals,eigvecs,kvectors,fermi,
            smearing=T)
# mean field
    mf,edc,charge,mag = magnetic_mean_field(voccs,U,collinear=collinear,
                                totkp=totkp)
    t3 = time.time()
    error = np.max(np.abs(old_mf-mf)) # absolute difference
    # total energy
    print("DC energy",edc)
    print("Occupied energies",np.sum(eoccs)/totkp)
    etot = np.sum(eoccs)/totkp + edc  # eigenvalues and double counting
    file_etot.write(str(ite)+"    "+str(etot)+"\n") # write energy in file
    file_error.write(str(ite)+"    "+str(error)+"\n") # write energy in file
    file_etot.flush()
    file_error.flush()
    totcharge = np.sum(charge).real # total charge
    avcharge = totcharge/nat # average charge
    htmp.write_magnetization() # write in a file
    if save: inout.save(mf,mf_file) # save the mean field
    ######
    if not silent:
      print("Times in diagonalization",t2-t1)
      print("Times in new mean field",t3-t2)
      print("\n")
      print("Iteration number =",ite)
      print("Error in SCF =",error)
      print("Fermi energy =",fermi)
      print("Total energy =",etot)
      print("Total charge =",totcharge)
      mx = htmp.extract(name="mx")
      my = htmp.extract(name="my")
      mz = htmp.extract(name="mz")
      print("Mag =",np.sum(mx),np.sum(my),np.sum(mz))
      print("Abs mag =",np.sum(np.abs(mx)),np.sum(np.abs(my)),np.sum(np.abs(mz)))
      print("Average charge =",avcharge)
    old_mf = old_mf*mix + mf*(1.-mix) # mixing
    if error<maxerror or os.path.exists("STOP"): # if converged break
      break
    if ite>=maxite: break # if too many iterations
  file_etot.close() # close file
  file_error.close() # close file
  # output result
  class scftmp(): pass # create an empty class
  scf = scftmp() # empty class
  scf.hamiltonian = htmp.copy() # copy Hamiltonian
  scf.hamiltonian.intra -= fermi*np.identity(ndim) # shift Fermi energy
  scf.total_energy = etot # store total energy
  scf.mf = mf # store mean field matrix
  scf.magnetization = mag # store magnetization
  scf.hamiltonian.write_magnetization() # write magnetization into a file
  return scf # return mean field




def selective_U_matrix(Us,m):
  """Return a mean field matrix, using a different U for each atom"""
  try: Us[0] # try to get an index
  except: return Us*m # return
  if len(Us)!=len(m)//2: raise
  mout = m.copy()
  fac = np.matrix(np.identity(len(Us)*2),dtype=np.complex)
  for i in range(len(Us)):
    fac[2*i,2*i] = Us[i]
    fac[2*i+1,2*i+1] = Us[i]
  return m*fac # return matrix



def selective_U_vector(Us,m):
  """Return a mean field matrix, using a different U for each atom"""
  try:
    Us[0] # try to get an index
    return np.array(Us)*m
  except: return Us*m # return






def magnetic_mean_field(wf,U,collinear=False,totkp=1):
  """Return the mean field matrix"""
  (vdup,vddn,vxc,ndn,nup,xc) = get_udxc(wf,totkp=totkp) # density and XC
  if collinear: mf = ndn + nup
  else: mf = ndn + nup - xc - xc.H
  mf = selective_U_matrix(U,mf.todense()) # new intramatrix mean field
  edc = -np.sum(selective_U_vector(U,vdup*vddn)).real
  edc += np.sum(selective_U_vector(U,vxc*np.conjugate(vxc))).real
  charge = (vdup+vddn).real # total charge
  mag = np.array([vxc.real,vxc.imag,(vdup-vddn)/2.]).transpose().real
  return mf,edc,charge,mag # return mean field and double counting energy











def spinless_mean_field(wf,U,collinear=False,totkp=1):
  """Return the mean field matrix"""
  raise # not finished



def get_udxc(voccs,weight=None,totkp=1):
  """Get up/down densities and corresponding mean field matrices"""
  ndim = voccs.shape[1] # dimension of the matrix
  if weight is not None:
    if len(weight)!=voccs.shape[0]: raise # inconsistent dimensions
  nat = ndim//2 # one half
  pdup = np.array([[2*i,2*i] for i in range(nat)]) # up density
  pddn = pdup + 1 # down density
  pxc = np.array([[2*i,2*i+1] for i in range(nat)]) # exchange
  if weight is None: # no weight
    vdup = correlatorsf90.correlators(voccs,pdup)/totkp
    vddn = correlatorsf90.correlators(voccs,pddn)/totkp
    vxc = correlatorsf90.correlators(voccs,pxc)/totkp
  else: # with weight
    vdup = correlatorsf90.correlators_weighted(voccs,weight,pdup)/totkp
    vddn = correlatorsf90.correlators_weighted(voccs,weight,pddn)/totkp
    vxc = correlatorsf90.correlators_weighted(voccs,pxc)/totkp
  ndn = csc_matrix((vdup,pddn.transpose()),dtype=np.complex,shape=(ndim,ndim))
  nup = csc_matrix((vddn,pdup.transpose()),dtype=np.complex,shape=(ndim,ndim))
  xc = csc_matrix((np.conjugate(vxc),pxc.transpose()),
                           dtype=np.complex,shape=(ndim,ndim))
  return (vdup,vddn,vxc,ndn,nup,xc) # return everything



def hubbardscf_spinless(h,g=1.0,nkp = 100,filling=0.5,mag=None,mix=0.9,
                  maxerror=1e-05,silent=False,mf=None,
                  smearing=None,collinear=False,fermi_shift=0.0):
  """ Solve a selfconsistent Hubbard mean field"""
  mix = 1. - mix
  U = g # redefine
  if h.has_spin: raise # only for spinless systems
  os.system("rm -f STOP") # remove stop file
  from scipy.linalg import eigh
  nat = h.intra.shape[0] # number of atoms
  htmp = h.copy()  # copy hamiltonian
  # generalate the necessary list of correlators
  if mf is None: # generate initial mean field
    mf = np.random.random(h.intra.shape) # random matrix
  else: old_mf = mf # use guess
  # get the pairs for the correlators
  ndim = h.intra.shape[0] # dimension
  totkp = nkp**(h.dimensionality) # total number of kpoints
  file_etot = open("SCF_ENERGY.OUT","w")
  file_error = open("SCF_ERROR.OUT","w")
  ite = 0 # iteration counter
  scf = scfclass(h) # create scf class
  while True: # infinite loop
    ite += 1 # increase counter
    htmp.intra = h.intra + old_mf # add mean field 
    t1 = time.time()
# get eigenvectors
    eigvals,eigvecs,kvectors = htmp.get_eigenvectors(nk=nkp,kpoints=True)
    eigvecs = np.conjugate(eigvecs)
# fermi energy
    t2 = time.time()
    fermi = get_fermi_energy(eigvals,filling,fermi_shift=fermi_shift)
# occupied states
    eoccs,voccs,koccs = get_occupied_states(eigvals,eigvecs,kvectors,fermi)
# spinless mean field
    mf,edc,charge = spinless_mean_field(voccs,U,totkp=totkp)
    t3 = time.time()
    print("Times in diagonalization",t2-t1)
    print("Times in new mean field",t3-t2)
    error = np.max(np.abs(old_mf-mf)) # absolute difference
    # total energy
    etot = np.sum(eoccs)/totkp + edc  # eigenvalues and double counting
    file_etot.write(str(ite)+"    "+str(etot)+"\n") # write energy in file
    file_error.write(str(ite)+"    "+str(error)+"\n") # write energy in file
    totcharge = np.sum(charge).real # total charge
    avcharge = totcharge/nat # average charge
    ######
    if not silent:
      print("\n")
      print("Iteration number =",ite)
      print("Error in SCF =",error)
      print("Fermi energy =",fermi)
      print("Total energy =",etot)
      print("Total charge =",totcharge)
      print("Average charge =",avcharge)
    old_mf = old_mf*mix + mf*(1.-mix) # mixing
    if error<maxerror or os.path.exists("STOP"): # if converged break
      break
    if ite>maxite: break # if too many iterations
  file_etot.close() # close file
  file_error.close() # close file
  # output result
  class scfclass(): pass # create an empty class
  scf = scftmp() # empty class
  scf.hamiltonian = htmp.copy() # copy Hamiltonian
  scf.hamiltonian.intra -= fermi*np.identity(ndim) # shift Fermi energy
  scf.total_energy = etot # store total energy
  scf.mf = mf # store mean field matrix
  scf.magnetization = mag
  scf.hamiltonian.write_magnetization() # write magnetization into a file



from __future__ import print_function
import scipy.sparse.linalg as lg
from scipy.sparse import csc_matrix
from scipy.sparse import coo_matrix,bmat
import time
import numpy as np
import os
import scipy.optimize as optimize
from . import klist
from . import inout
from . import algebra
from . import meanfield
from . import groundstate
from . import parallel

from .meanfield import guess # function to calculate guesses
try:
  from . import correlatorsf90
  multicorrelator = correlatorsf90.multicorrelator
  multicorrelator_bloch = correlatorsf90.multicorrelator_bloch
  use_multicorrelator = True
except:
#  print("WARNING, FORTRAN not working in scftypes.py")
  use_multicorrelator = False
#  raise
#  def multicorrelator


timing = True


def load(input_file="SCF.pkl"):
  """Get a selfconsistent calculation"""
  return inout.load(input_file)

class scfclass(): 
  """Class for a selfconsistent calculation"""
  def copy(self):
    from copy import deepcopy
    return deepcopy(self)
  def save(self,output_file="SCF.pkl"):
    inout.save(self,output_file)
  def __init__(self,h):
    self.has_index2vec = False
    self.iteration = 0 # initialize
    self.scfmode = "filling" # do SCF fixing the number of filled states
    self.interactions = [] # empty list
    self.nkgrid = 1 # one kpoint
    self.mf = dict() # dictionary
    self.kfac = 1.0 # normalization factor for the number of kpoints
    self.fermi = 0.0 # fermi energy
    self.error = 0.0 # error in the selfconsistency
    self.silent = False
    self.gap = 0.0 # gap of the system
    self.smearing = None
    self.write_vabv = False # write EV in each iteration
    if use_multicorrelator: # fortran library is present
      self.correlator_mode = "multicorrelator" # multicorrelator mode 
    else:
      self.correlator_mode = "plain" # multicorrelator mode 
#    self.correlator_mode = "1by1" # multicorrelator mode 
    self.bloch_multicorrelator = False # multicorrelator with Bloch phases
    self.enforce_symmetry = "none" # do not enforce any symmetry
    self.hamiltonian0 = h.get_multicell() # store Hamiltonian
    self.hamiltonian = self.hamiltonian0.copy() # store Hamiltonian
    self.has_eh = self.hamiltonian.has_eh # has electron-hole
    self.is_sparse = self.hamiltonian.is_sparse # sparse Hamiltonian
    self.sites = len(self.hamiltonian.geometry.r) # number of sites
    self.iteration = 0 # first iteration
    self.filling = 0.5 # half filling
    self.mixing = 0.9 # initialize the mixing
    self.fermi_shift = 0.0 # shift in the fermi energy
    self.energy_cutoff = 1.0 # energy for the selfconsistency
    self.num_waves = 10 # number of waves to compute
    self.use_weights = False # calculate SCF using weights
  def update_occupied_states(self,fermi_shift=0.0):
    """Get the eigenvectors for a mesh of kpoints"""
    mine = None # minimum energy
    if self.scfmode=="fermi" and self.is_sparse:
      self.hamiltonian.turn_sparse()
      print("WARNING!!! using sparse mode")
      print("Use this mode only if you know what you are doing!!!!\n\n")
      es,ws,ks = self.hamiltonian.get_eigenvectors(nk=self.nkgrid,kpoints=True,
                              sparse=True,numw=self.num_waves)
      if np.max(np.abs(es))*0.9<self.energy_cutoff: 
        print("NOT ENOUGH STATES, recalling with",self.num_waves*2)
        self.num_waves += 5 
        self.update_occupied_states() # call again
#  raise 
    else: # any other, use brute force
      es,ws,ks = self.hamiltonian.get_eigenvectors(nk=self.nkgrid,kpoints=True)
#    self.kfac = float(len(ks))/self.hamiltonian.intra.shape[0] 
    # number of kpoints of the calculation
#    mine = min(es)*0.9 # minimum energy retained
    self.kfac = len(klist.kmesh(self.hamiltonian.dimensionality,nk=self.nkgrid))
    if self.scfmode=="filling": # get the fermi energy if the mode requires it 
      self.fermi = get_fermi_energy(es,self.filling,fermi_shift=fermi_shift) 
    elif self.scfmode=="fermi": pass # do nothing
    self.gap = get_gap(es,self.fermi) # store the gap
    if self.energy_cutoff is not None:
      if self.gap>self.energy_cutoff/2: 
        print("\nEnergy cutoff is to small, halting",self.gap,"\n\n")
        raise
      print("Warning!!!! Performing calculation with an energy cutoff")
    eoccs,voccs,koccs = get_occupied_states(es,ws,ks,self.fermi,
                            mine=self.energy_cutoff,smearing = self.smearing)
    self.wavefunctions = voccs # store wavefunctions
#    print(len(voccs),voccs.shape,len(voccs[0]))
    self.energies = eoccs # store energies
    self.kvectors = koccs # store kvectors
  def update_occupied_states_kpoint(self,k=[0.,0.,0.]):
    """Update the object with the eigenvectors of a single kpoint,
    this function is used for the adaptive method of SCF"""
    if not self.scfmode=="fermi": raise # only for this mode
    es,ws,ks = self.hamiltonian.get_eigenvectors(nk=self.nkgrid,
            kpoints=True,k=k)
    self.kfac = 1
    self.gap = get_gap(es,self.fermi) # store the gap
    eoccs,voccs,koccs = get_occupied_states(es,ws,ks,self.fermi) # occupied states
    self.wavefunctions = voccs # store wavefunctions
    self.energies = eoccs # store energies
    self.kvectors = koccs # store kvectors
  def adaptive_correlator(self):
    """Calculate the correlators using the Simpson's adaptive algorithm"""
    raise # not finished yet
  def update_hamiltonian(self):
    """Updates the total mean field Hamiltonian"""
    self.hamiltonian = self.hamiltonian0.copy() # copy original
    self.hamiltonian.intra += self.mf[(0,0,0)] # add mean field
    if self.hamiltonian.dimensionality>0:
      from . import multicell
      hopping = self.hamiltonian.hopping # empty list
      for key in self.mf: # loop over the mean field
        if key==(0,0,0): continue
        m = self.mf[key] # get this hopping
        hopping.append(multicell.Hopping(d=key,m=m)) # add this one
      self.hamiltonian.hopping = multicell.collect_hopping(self.hamiltonian)
  def setup_interaction(self,mode="Hubbard",g=1.0,**kwargs):
    """Create the operators that will determine the interactions"""
    if timing: t0 = time.perf_counter()
    interactions = [] # empty list
    has_eh = self.hamiltonian.has_eh # has electron hole
    nat = self.sites # number of sites
#    self.correlator_mode = "1by1" # mode to calculate the correlators
    if has_eh: # for superconducting systems
      self.scfmode = "fermi"  # SCF calculation fixing the fermi energy
      self.fermi = 0.0 # set the fermi energy in 0
      if mode=="Hubbard" or mode=="U": # Hubbard model
        print("Adding Hubbard interaction")
        for i in range(nat): 
          interactions.append(meanfield.hubbard_pairing_ud(i,nat,g=g)) 
          interactions.append(meanfield.hubbard_pairing_du(i,nat,g=g)) 
      elif mode=="V": # V interaction
        print("Adding V interaction")
        self.bloch_multicorrelator = True
        from . import neighbor
 #       self.nkgrid = 1 # only implemented in Gamma point
        for d in self.hamiltonian.geometry.neighbor_directions(): # loop
          ri = self.hamiltonian.geometry.r # positions
          rj = self.hamiltonian.geometry.replicas(d) # positions
          ijs = neighbor.find_first_neighbor(ri,rj) # return pairs
          if len(ijs)>0: self.mf[tuple(d)] = self.hamiltonian.intra*0. # initialize
          for (i,j) in ijs: # loop over pairs of sites
#            interactions.append(meanfield.v_pairing_du(i,j,nat,g=g,d=d)) 
#            interactions.append(meanfield.v_pairing_du(j,i,nat,g=g,d=d)) 
            interactions.append(meanfield.v_pairing_dd(i,j,nat,g=g,d=d,channel="ee")) 
            interactions.append(meanfield.v_pairing_dd(i,j,nat,g=g,d=d,channel="hh")) 
            interactions.append(meanfield.v_pairing_uu(i,j,nat,g=g,d=d,channel="ee")) 
            interactions.append(meanfield.v_pairing_uu(i,j,nat,g=g,d=d,channel="hh")) 
    else: # no electron hole symmetry
      if mode=="Hubbard" or mode=="U": # Hubbard model
        if self.hamiltonian.has_spin: # spinful
          print("Adding Spinful Hubbard interaction")
          for i in range(nat): 
            interactions.append(meanfield.hubbard_density(i,nat,g=g)) 
            interactions.append(meanfield.hubbard_exchange(i,nat,g=g)) 
        else:
          print("Adding spinless Hubbard interaction")
          for i in range(nat): 
            interactions.append(meanfield.spinless_hubbard_density(i,nat,g=g)) 
      elif mode=="Hubbard collinear": # Hubbard model
        print("Adding Hubbard collinear interaction")
        for i in range(nat): 
          interactions.append(meanfield.hubbard_density(i,nat,g=g)) 
  # store this interaction
      elif mode=="Coulomb": # Coulomb interaction
        interactions = meanfield.coulomb_interaction(
                self.hamiltonian.geometry,has_spin=self.hamiltonian.has_spin,
                vc=g,**kwargs)
      elif mode=="fastCoulomb": # Coulomb interaction
        interactions = meanfield.fast_coulomb_interaction(
                self.hamiltonian.geometry,has_spin=self.hamiltonian.has_spin,
                vc=g,**kwargs)
      elif mode=="V": # V interaction
#        self.correlator_mode = "1by1" # mode to calculate the correlators
        self.bloch_multicorrelator = True
        from . import neighbor
#        self.nkgrid = 1 # only implemented in Gamma point
        for d in self.hamiltonian.geometry.neighbor_directions(): # loop
          ri = self.hamiltonian.geometry.r # positions
          rj = self.hamiltonian.geometry.replicas(d) # positions
          ijs = neighbor.find_first_neighbor(ri,rj) # return pairs
          if len(ijs)>0: self.mf[tuple(d)] = 0.0j # initialize
          for (i,j) in ijs: # loop over pairs of sites
            if self.hamiltonian.has_spin: # spinful
              interactions.append(meanfield.v_ij(i,j,nat,g=g,d=d,spini=0,spinj=0)) 
              interactions.append(meanfield.v_ij(i,j,nat,g=g,d=d,spini=0,spinj=1)) 
              interactions.append(meanfield.v_ij(i,j,nat,g=g,d=d,spini=1,spinj=0)) 
              interactions.append(meanfield.v_ij(i,j,nat,g=g,d=d,spini=1,spinj=1)) 
            else: # spinless
              # factor 2 in g due to spin degree of freedom
              interactions.append(meanfield.v_ij_spinless(i,j,nat,g=2*g,d=d)) 
      else: raise # ups
    self.interactions += interactions # store list
    if timing: print("Time in creating MF operators",time.perf_counter()-t0)
    self.setup_multicorrelator()
  def setup_multicorrelator(self):
    """Create the neccesary arrays to calculate the different
    correlators efficiently"""
    lamb = [] # empty list
    ijk = []
    dv = [] # direction of the interaction
    k = 0
    for v in self.interactions: # loop
      ac = coo_matrix(v.a)       
      bc = coo_matrix(v.b)       
      for (i,j,d) in zip(ac.row,ac.col,ac.data): 
        ijk.append([i,j,k])
        lamb.append(d) # store data
        dv.append(v.dir) # store direction
      k += 1 # increase counter
      for (i,j,d) in zip(bc.row,bc.col,bc.data): 
        ijk.append([i,j,k])
        lamb.append(d) # store data
        dv.append(-np.array(v.dir)) # store direction, be aware of the sign!!!
      k += 1 # increase counter
    ijk = np.array(ijk,dtype=np.int)
    lamb = np.array(lamb,dtype=np.complex) # data array
    self.ijk = ijk # first array
    self.lamb = lamb # data array
    self.dir = np.array(dv,dtype=np.int) # data array
#    self.tensormf = algebra.sparsetensor.Tensor3(ijk[:,0],
#            ijk[:,1],ijk[:,2],lamb,
#            shape=(v.a.shape[0],v.a.shape[0],k))
  def update_expectation_values(self):
    """Calculate the expectation values of the different operators"""
    # this conjugate comes from being inconsistent
    # in the routines to calculate exectation values
    voccs = np.array(np.conjugate(self.wavefunctions)) # get wavefunctions
    ks = self.kvectors # kpoints
#    self.correlator_mode = "plain"
    mode = self.correlator_mode # 
    if mode=="plain": # conventional mode
        plain_expectation_value(self)
#      for v in self.interactions:
#        v.vav = np.trace(voccs@v.a@np.conjugate(voccs).T)/self.kfac # <vAv>
#        v.vbv = np.trace(voccs@v.b@np.conjugate(voccs).T)/self.kfac # <vBv>
    elif mode=="1by1": # conventional mode
      for v in self.interactions:
        phis = [self.hamiltonian.geometry.bloch_phase(v.dir,k*0.) for k in ks]
        v.vav = meanfield.expectation_value(voccs,v.a,np.conjugate(phis))/self.kfac # <vAv>
        v.vbv = meanfield.expectation_value(voccs,v.b,phis)/self.kfac # <vBv>
      self.v2cij() # update the v vector
    elif mode=="multicorrelator": # multicorrelator mode
      numc = len(self.interactions)*2 # number of correlators
      if self.bloch_multicorrelator:
        cs = multicorrelator_bloch(voccs,ks,self.lamb,self.ijk,self.dir,numc)
      else: cs = multicorrelator(voccs,self.lamb,self.ijk,numc)
      self.cij = cs/self.kfac # store in the object, already normalized
      self.cij2v() # update the expectation values
    else: raise
  def cij2v(self):
    """Update the values of vav and vbv using the values of cij"""
    cs = self.cij # get the array
    k = 0
    for v in self.interactions:
      v.vav = cs[k]
      k += 1
      v.vbv = cs[k]
      k += 1
  def v2cij(self):
    """Update the value of cij with the values of v"""
    cs = np.zeros(len(self.interactions)*2,dtype=np.complex) # get the array
    k = 0
    for v in self.interactions:
      cs[k] = v.vav
      k += 1
      cs[k] = v.vbv 
      k += 1
    self.cij = cs.copy() # store
  def update_mean_field(self,**kwargs):
    """Calculate the expectation values of the different operators"""
    from .scftk.updatescf import update_mean_field
    update_mean_field(self,**kwargs)
  def get_total_energy(self):
    """Return the total energy"""
    eout = np.sum(self.energies)/self.kfac 
    for v in self.interactions:
      if v.contribution=="AB": eout += -v.vav*v.vbv*v.g
      else: eout += -v.vav*v.vbv*v.g/2.
    self.total_energy = eout.real # store the total energy
    return eout.real # return total energy of the system
  def extract(self):
    """Write in file different parts of the Hamiltonian"""
    if self.hamiltonian.has_eh: groundstate.swave(self.hamiltonian) 
  def iterate(self):
    """Perform a single iteration"""
    if not self.silent: print("## Iteration number ",self.iteration)
    mixing = self.mixing 
    self.update_hamiltonian() # update the Hamiltonian
    try: self.hamiltonian.check() # check that nothing weird happened
    except: 
        print("WARNING, hamiltonian is non hermitian in mean field")
        raise
    t1 = time.perf_counter()
    self.update_occupied_states(fermi_shift=self.fermi_shift)
    t2 = time.perf_counter()
    self.update_expectation_values() # calculate expectation values
    t3 = time.perf_counter()
    self.update_mean_field(mixing=mixing) # calculate mean field
    t4 = time.perf_counter()
    self.iteration += 1
    if not self.silent:
      print("Time in diagonalization",t2-t1)
      print("Time in expectation values",t3-t2)
      print("Time in new Hamiltonian",t4-t3)
      print("\n")
  def solve(self):
    """Solve the selfconsistent problem"""
    return
    self.iterate() # first iteration
    def r2z(v): # convert real to complex
      n = v.shape[0]//2
      return v[0:n] + 1j*v[n:2*n]
    def z2r(v): # convert complex to real
      return np.concatenate([v.real,v.imag])

    def fopt(cij): # function to return the error
      self.cij = r2z(cij) # store this vector
      self.cij2v()  # update the v vectors
      self.iterate() # do an iteration
      print(self.cij)
      self.v2cij() # convert the v to cij
      return z2r(self.cij)-cij
    cij = optimize.broyden1(fopt,z2r(self.cij),f_tol=1e-8,max_rank=10)
#    cij = optimize.fsolve(fopt,z2r(self.cij),xtol=1e-8)
#    cij = optimize.anderson(fopt,z2r(self.cij),f_tol=1e-6,w0=0.1)
#    cij = optimize.newton_krylov(fopt,z2r(self.cij),f_tol=1e-6,outer_k=8)
    self.cij = cij
    self.cij2v() # update




def directional_mean_field(vecs):
  """ Creates an initial mean field accoring to certain vectors"""
  from .hamiltonians import sx,sy,sz
  mf = [[None for i in vecs] for j in vecs] # create
  for i in range(len(vecs)):
    v = vecs[i]
    mf[i][i] = v[0]*sx + v[1]*sy + v[2]*sz # add contribution
  return bmat(mf).todense()



def get_occupied_states(es,ws,ks,fermi,smearing=None,mine=None):
  """Return the occupied states"""
  if smearing is None: # no smearing
    voccs = [] # occupied vectors
    eoccs = [] # occupied eigenvalues
    koccs = [] # occupied eigenvalues
    if mine is None: mine = -1000000 # accept all
    else: mine = -np.abs(mine)
    for (e,v,k) in zip(es,ws,ks): # loop over eigenvals,eigenvecs
      if mine<e<fermi:  # if level is filled, add contribution
        voccs.append(v) # store
        eoccs.append(e) # store
        koccs.append(k) # store
    voccs = np.array(voccs)  # as array
    eoccs = np.array(eoccs)  # as array
    koccs = np.array(koccs)  # as array
  else:
    voccs = [] # occupied vectors
    eoccs = [] # occupied eigenvalues
    koccs = [] # occupied eigenvalues
    if mine is None: mine = -1000000 # accept all
    else: mine = -np.abs(mine)
    for (e,v,k) in zip(es,ws,ks): # loop over eigenvals,eigenvecs
      weight = 1./(np.exp((e-fermi)/smearing)+1.0) # occupation
      voccs.append(v*weight) # store
      eoccs.append(e*weight) # store
      koccs.append(k) # store
    voccs = np.matrix(np.array(voccs))  # as array
    eoccs = np.array(eoccs)  # as array
    koccs = np.array(koccs)  # as array
  return eoccs,voccs,koccs


from .spectrum import get_fermi_energy


def get_gap(es,fermi):
  """Return the gap"""
  dife = es - fermi # difference in energy
  gap = np.min(np.abs(dife))*2 # gap
  return gap





from .selfconsistency.hubbard import hubbardscf
from .selfconsistency.coulomb import coulombscf
from .selfconsistency.densitydensity import densitydensity

repulsive_hubbard = hubbardscf
from .selfconsistency.attractive_hubbard_spinless import attractive_hubbard

def get_super_correlator(voccs,weight=None,totkp=1):
  """Get the different correlators for a superconducting system"""
  from . import correlatorsf90
  ndim = voccs.shape[1] # dimension of the matrix
  if weight is not None:
    if len(weight)!=voccs.shape[0]: raise # inconsistent dimensions
  nat = ndim//4 # one fourth
  corrs = [] # empty list with the different correlators
  ops = [] # list with the operators
  for i in nat: # loop over atoms
    corrs.append([4*i,4*i]) # up density
    ops.append(csc_matrix(([1.],[1,1]),dtype=np.complex)) # down operator
    corrs.append([4*i+1,4*i+1]) # up density
    ops.append(csc_matrix(([1.],[0,0]),dtype=np.complex)) # down operator
    corrs.append([4*i,4*i+1]) # up density
    ops.append(csc_matrix(([-1.],[1,0]),dtype=np.complex)) # down operator
    corrs.append([4*i+1,4*i]) # up density
    ops.append(csc_matrix(([-1.],[0,1]),dtype=np.complex)) # down operator
  pdup = np.array([[4*i,4*i] for i in range(nat)]) # up density
  pddn = pdup + 1 # down density
  pxc = np.array([[4*i,4*i+1] for i in range(nat)]) # exchange
  deltadd = np.array([[4*i+1,4*i+2] for i in range(nat)]) # Delta dd
  deltauu = np.array([[4*i,4*i+3] for i in range(nat)]) # Delta uu
  deltaud = np.array([[4*i,4*i+2] for i in range(nat)]) # Delta ud
  deltadu = np.array([[4*i+1,4*i+3] for i in range(nat)]) # Delta du
  if weight is None: # no weight
    vdup = correlatorsf90.correlators(voccs,pdup)/totkp
    vddn = correlatorsf90.correlators(voccs,pddn)/totkp
    vxc = correlatorsf90.correlators(voccs,pxc)/totkp
    vdeltadd = correlatorsf90.correlators(voccs,deltadd)/totkp
    vdeltauu = correlatorsf90.correlators(voccs,deltauu)/totkp
    vdeltadu = correlatorsf90.correlators(voccs,deltadu)/totkp
    vdeltaud = correlatorsf90.correlators(voccs,deltaud)/totkp
  else: # with weight
    raise
    vdup = correlatorsf90.correlators_weighted(voccs,weight,pdup)/totkp
    vddn = correlatorsf90.correlators_weighted(voccs,weight,pddn)/totkp
    vxc = correlatorsf90.correlators_weighted(voccs,pxc)/totkp
  ndn = csc_matrix((vdup,pddn.transpose()),dtype=np.complex,shape=(ndim,ndim))
  nup = csc_matrix((vddn,pdup.transpose()),dtype=np.complex,shape=(ndim,ndim))
  xc = csc_matrix((np.conjugate(vxc),pxc.transpose()),
                           dtype=np.complex,shape=(ndim,ndim))
  return (vdup,vddn,vxc,ndn,nup,xc) # return everything





def write_magnetization(mag):
  """Write magnetization in a file"""
  fo = open("MAGNETIZATION.OUT","w")
  ix = 1
  for m in mag:
    fo.write(str(ix)+"  ")
    fo.write(str(m[0])+"  ")
    fo.write(str(m[1])+"  ")
    fo.write(str(m[2])+"\n")
    ix += 1
  fo.close()






#
#def selfconsistency(h,g=1.0,nkp = 100,filling=0.5,mix=0.2,
#                  maxerror=1e-05,silent=False,mf=None,nk=None,
#                  smearing=None,fermi_shift=0.0,save=True,
#                  mode="Hubbard",energy_cutoff=None,maxite=1000,
#                  broyden=False,callback=None,**kwargs):
#  """ Solve a generalized selfcnsistent problem"""
#  if nk is not None: nkp = nk # redefine
#  mf_file = "MF.pkl"
#  os.system("rm -f STOP") # remove stop file
#  nat = h.intra.shape[0]//2 # number of atoms
#  htmp = h.copy()  # copy hamiltonian
#  htmp.turn_dense() # turn to dense Hamiltonian
#  # generalate the necessary list of correlators
#  if mf is None: # generate initial mean field
#    try:  old_mf = inout.load(mf_file) # load the file
#    except: old_mf = meanfield.guess(h,"random") # random guess
#  else: old_mf = mf # use guess
#  # get the pairs for the correlators
#  ndim = h.intra.shape[0] # dimension
#  totkp = nkp**(h.dimensionality) # total number of kpoints
#  file_etot = open("SCF_ENERGY.OUT","w")
#  file_gap = open("SCF_GAP.OUT","w")
#  file_error = open("SCF_ERROR.OUT","w")
#  ite = 0 # iteration counter
#  scf = scfclass(htmp) # create scf class
#  scf.nkgrid = nkp
#  scf.silent = silent
#  scf.mixing = mix
#  scf.mode = mode # store the mode
#  scf.g = g # coupling
#  scf.smearing = smearing
#  scf.energy_cutoff = energy_cutoff # energy_cutoff
#  scf.filling = filling # filling of the system
##  scf.mf0 = old_mf # initial mean field
#  if type(mode) is type(dict()): # of mode is a dictionary add several interactions
#    print("Adding multiple interactions")
#    for key in mode:
#      scf.setup_interaction(g=mode[key],mode=key,**kwargs) # create the interaction matrices
#  else: # conventional way
#    scf.setup_interaction(g=g,mode=mode,**kwargs) # create the interaction matrices
#  if type(old_mf) is type(dict()): # of mode is a dictionary add several interactions
#    scf.mf = old_mf # initial mean field
#  else:
#    scf.mf[(0,0,0)] = old_mf # initial mean field
##  scf.solve()
#  stop_scf = False # do not stop
#  scf.maxerror = maxerror
##  print("BEGINNING OF BROYDEN")
##  print("END OF BROYDEN")
#  while True: # infinite loop
#    scf.iterate() # do an iteration
#    if scf.hamiltonian.has_spin: 
#        scf.hamiltonian.write_magnetization() # write the magnetization
#    eout = scf.get_total_energy() # total energy
#    if not silent: print("Total energy",eout)
##    etot = np.sum(eoccs)/totkp + edc  # eigenvalues and double counting
#    ite = scf.iteration
#    file_etot.write(str(ite)+"    "+str(scf.get_total_energy())+"\n") 
#    file_error.write(str(ite)+"    "+str(scf.error)+"\n") # write energy in file
#    file_gap.write(str(ite)+"    "+str(scf.gap)+"\n") # write energy in file
##    totcharge = np.sum(charge).real # total charge
##    avcharge = totcharge/nat # average charge
##    ######
#    if callback is not None: callback(scf.hamiltonian) # callback
#    if not silent:
#      print("\n")
#      print("Error in SCF =",scf.error)
#      print("Fermi energy =",scf.fermi)
#      print("Gap =",scf.gap)
#    if scf.error<maxerror or os.path.exists("STOP") or scf.iteration==maxite: 
#      stop_scf = True # stop the calculation after the next iteration
#      scf.mixing = 1.0 # last iteration with mixing one
#      scf.smearing = None # last iteration without smearing
#    if stop_scf: break # stop the calculation
#  if broyden: scf = meanfield.broyden_solver(scf)
#  if save: inout.save(scf.mf,mf_file) # save mean field
#  file_etot.close() # close file
#  file_error.close() # close file
#  file_gap.close() # close file
#  if scf.scfmode=="filling":  scf.hamiltonian.shift_fermi(-scf.fermi)
#  return scf # return mean field




def plain_expectation_value(self):
    """Compute the expectation values using the plain scheme"""
    voccs = np.array(np.conjugate(self.wavefunctions))
    if parallel.cores==1: # single core execution
      for v in self.interactions:
        v.vav = np.trace(voccs@v.a@np.conjugate(voccs).T)/self.kfac # <vAv>
        v.vbv = np.trace(voccs@v.b@np.conjugate(voccs).T)/self.kfac # <vBv>
    else: # multicore execution
        ma = [v.a for v in self.interactions]
        mb = [v.b for v in self.interactions]
        kfac = self.kfac
        def fun(i):
            vav = np.trace(voccs@ma[i]@np.conjugate(voccs).T)/kfac
            vbv = np.trace(voccs@mb[i]@np.conjugate(voccs).T)/kfac
            return [vav,vbv]
        # compute expectation values
        vv = parallel.pcall(fun,range(len(self.interactions)))
        for i in range(len(self.interactions)):
            self.interactions[i].vav = vv[i][0]
            self.interactions[i].vbv = vv[i][1]


# workaround
from .meanfield import Vinteraction as selfconsistency




# routines to extract channels from a matrix
from __future__ import division
import numpy as np
from . import algebra
from .htk.matrixcomponent import spin_mixing_part
from . import superconductivity

def spin_channel(m,spin_column=None,spin_row=None,has_spin=True):
  """Extract a channel from a matrix"""
  if not has_spin: return m # return initial
  if (spin_row is None) or (spin_column is None): return m # return initial
  n = m.shape[0] # shape of the matrix
  n2 = n//2 # number of orbitals
  out = np.zeros((n,n),dtype=np.complex)
  if spin_column=="up": ii = 0
  else: ii = 1
  if spin_row=="up": jj = 0
  else: jj = 1
  for i in range(n2):
    for j in range(n2): out[i,j] = m[2*i+ii,2*j+jj]
  return np.matrix(out)



def swave(m):
  """Extract the swave pairing from a matrix, assuming
  the Nambu spinor basis"""
  n = m.shape[0]//4 # number of sites
  ds = np.zeros(n,dtype=np.complex) # pairing
  for i in range(n):
    ds[i] = m[4*i,4*i+2] # get the pairing
  return ds




def mz(m):
  """Extract the z component of the magnetism, assume spin degree of freedom"""
  n = m.shape[0]//2 # number of sites
  ds = np.zeros(n).real # pairing
  for i in range(n):
    ds[i] = -(m[2*i+1,2*i+1] - m[2*i,2*i]).real/2. # get the pairing
  return ds



def mx(m):
  """Extract the z component of the magnetism, assume spin degree of freedom"""
  n = m.shape[0]//2 # number of sites
  ds = np.zeros(n).real # pairing
  for i in range(n):
    ds[i] = m[2*i,2*i+1].real
  return ds



def my(m):
  """Extract the z component of the magnetism, assume spin degree of freedom"""
  n = m.shape[0]//2 # number of sites
  ds = np.zeros(n).real # pairing
  for i in range(n):
    ds[i] = -m[2*i,2*i+1].imag
  return ds



def onsite(m,has_spin=True):
  """Extract the onsite energy"""
  if has_spin: # has spin degree of freedom
    n = m.shape[0]//2 # number of sites
    ds = np.zeros(n).real # pairing
    for i in range(n):
      ds[i] = (m[2*i,2*i].real + m[2*i+1,2*i+1].real)/2.
    return ds
  else:
    n = m.shape[0] # number of sites
    ds = np.zeros(n).real # pairing
    for i in range(n):
      ds[i] = m[i,i].real
    return ds



def hopping_spinful(m,cutoff=0.001):
  """Extract hopping"""
  n = m.shape[0]//2 # number sites
  ii = []
  jj = []
  ts = []
  for i in range(n):
    for j in range(i,n):
      t = (np.abs(m[2*i,2*j]) + np.abs(m[2*i+1,2*j+1]))/2.
      t = (np.abs(m[2*i,2*j+1]) + np.abs(m[2*i+1,2*j]))/2.
      if t>cutoff:
        ii.append(i)
        jj.append(j)
        ts.append(t)
  return ii,jj,np.array(ts) # return pairs


def hopping_spinful_difference(m,cutoff=0.001,skip_same_site=False):
  """Extract hopping"""
  n = m.shape[0]//2 # number sites
  ii = []
  jj = []
  ts = []
  for i in range(n):
    for j in range(i,n):
      if i==j and skip_same_site: continue
      t = np.abs(m[2*i,2*j]) - np.abs(m[2*i+1,2*j+1])
      if abs(t)>cutoff:
        ii.append(i)
        jj.append(j)
        ts.append(t)
  return ii,jj,np.array(ts) # return pairs





def hopping_spinless(m,cutoff=0.001):
  """Extract hopping"""
  from scipy.sparse import coo_matrix
  m = coo_matrix(m) # transform to coo_matrix
  m.eliminate_zeros() # remove zeros
  row,col,data = m.row,m.col,m.data
  absd = np.abs(data) # absolute value
  row = row[absd>cutoff]
  col = col[absd>cutoff]
  data = data[absd>cutoff]
  return row,col,data




  
def extract_from_hamiltonian(self,name):
    """Extract a quantity from a Hamiltonian"""
    h0 = self.copy()
    if self.is_sparse: h0.turn_dense() # turn into dense form
    if name=="density":
      if self.has_eh: 
          h0.remove_nambu()
          m = h0.intra
      else: m = self.intra
      return onsite(m,has_spin=self.has_spin)
    elif name=="mx" and self.has_spin:
      if self.has_eh: h0.remove_nambu() # not implemented
      return mx(h0.intra)
    elif name in ["swave","SC"]:
        if self.check_mode("spinful_nambu"): 
            return swave(self.intra)
        elif self.check_mode("spinless_nambu"): 
            from .sctk import spinless
            return spinless.extract_swave(self.intra)
        else: raise
    elif name=="CDW":
        if self.geometry.has_sublattice: # if it has sublattice
            v = self.extract("density")
            v = v - np.mean(v) # remove average
            return v*np.array(self.geometry.sublattice)
    elif name=="my" and self.has_spin:
      if self.has_eh: h0.remove_nambu() # not implemented
      return my(h0.intra)
    elif name=="mz" and self.has_spin:
      if self.has_eh: h0.remove_nambu() # not implemented
      return mz(h0.intra)
    elif name=="spin_mixing":
        return extract_spin_mixing(self)
    elif name=="hopping_spin_mixing":
        return extract_hopping_spin_mixing(self)
    elif name=="superfluidity": # extract the absolute value
        if self.has_eh:
            from .superconductivity import dict2absdeltas
            (uu,dd,ud) = dict2absdeltas(self.get_multihopping().get_dict())
            return uu+dd+ud
    else: raise



def extract_onsite_matrix_function(h,**kwargs):
    """Extract a certain function"""
    h = h.copy() # copy
    if h.check_mode("spinful"): # not implemented
      m = h.intra # get the matrix
      n = len(h.geometry.r) # number of sites
      if 2*n!=m.shape[0]: raise
      def f(r):
        ind = h.geometry.get_index(r,**kwargs) # get the index
        if ind is None: return np.zeros((2,2),dtype=np.complex)
        else: return m[2*ind:2*ind+2,2*ind:2*ind+2]
    return f # return function

def extract_magnetism_function(h,**kwargs):
    """Function that return the magnetization"""
    fm = extract_onsite_matrix_function(h,**kwargs) # create the function
    def f(r):
        m = fm(r) # get the matrix
        mx = m[0,1].real
        my = m[0,1].imag
        mz = (m[0,0] - m[1,1]).real/2.
        return np.array([mx,my,mz])
    return f # return function


def extract_onsite_function(h,**kwargs):
    """Function that return the onsite energy"""
    fm = extract_onsite_matrix_function(h,**kwargs) # create the function
    def f(r):
        m = fm(r) # get the matrix
        return (m[0,0] + m[1,1]).real/2.
    return f # return function


def extract_spin_mixing(h):
    """Extract the spin mixing part of a Hamiltonian"""
    h = h.copy()
    h.remove_nambu() # remove nambu
    if not h.has_spin: raise
    dt = h.get_dict() # get the multihopping object
    out = 0 # output
    for key in dt: # loop
        m = dt[key] # get the matrix
        m = spin_mixing_part(m) # spin mixing part
        m = np.abs(np.array(algebra.todense(m))) # absolute value
        m = m*m # square value
        m = np.mean(m,axis=0) # sum over the first axis
        out = out + m # add to the output
    return out # return the mixing


def extract_hopping_spin_mixing(h):
    """Extract the spin mixing part of a Hamiltonian"""
    h = h.copy()
    h.remove_nambu() # remove nambu
    if not h.has_spin: raise
    dt = h.get_dict() # get the multihopping object
    out = 0 # output
    for key in dt: # loop
        m = dt[key] # get the matrix
        m = spin_mixing_part(m) # spin mixing part
        if key==(0,0,0): # discard onsite terms
            for i in range(m.shape[0]): m[i,i] = 0.0 # set to zero
        m = np.abs(np.array(algebra.todense(m))) # absolute value
        m = m*m # square value
        m = np.mean(m,axis=0) # sum over the first axis
        out = out + m # add to the output
    return out # return the mixing







import numpy as np

def target_moduli(v1,v2,m):
    """Given two vectors, obtain a linear combination with a certain moduli"""
    n = 2*int(m) # estimate of the number
    n = 1
    for i in range(-n,n+1):
        for j in range(-n,n+1):
            v = i*v1 + j*v2
            if abs(np.sqrt(v.dot(v))-m)<1e-3: return v
    raise # not found

def closest_kreplica(g,v,v0):
    """Return the closest replica of a kpoint close to another one"""
    if g.dimensionality!=2: raise # only for 2d so far
    w = v[0]*g.b1 + v[1]*g.b2 # redefine
    w0 = v0[0]*g.b1 + v0[1]*g.b2 # redefine
    n = 2
    dmax = 1e8
    for i in range(-n,n+1):
        for j in range(-n,n+1):
            wt = w + g.b1*i + g.b2*j # compute replica
            dw = wt - w0 # distance to reference point
            dw = np.sqrt(dw.dot(dw)) # distance
            if dw<dmax: 
                vo = v + np.array([i,j,0.])
                dmax = dw
    return vo # return vo



def closest_path(g,kps):
    """Given a list of kpoints, return the closest path"""
    out = [kps[0]] # first point
    for i in range(len(kps)-1): # loop over kpoints
        k = closest_kreplica(g,kps[i+1],out[i])
        out.append(k) # store
    return out



def k2path(g,kp,nk=100):
    """Given a set of kpoints, return the path"""
    kp = closest_path(g,kp) # reduced path
    kp = [np.array(k) for k in kp] # convert to arrays
    ks = [] # empty list of kpoints
    for i in range(len(kp)-1): # loop over pairs
      dk0 = kp[i+1] - kp[i] # difference
      dk = dk0[0]*g.b1 + dk0[1]*g.b2 # real reciprocal vector
      dk2 = dk.dot(dk) # moduli real vector
      nk2 = int(nk*np.sqrt(dk2/g.b1.dot(g.b1))) # in units of reciprocal vector
      steps = np.linspace(0.,1.,nk2,endpoint=False) # number of points
      for s in steps:
        ks += [kp[i] + dk0*s] # add kpoint
    return ks




# library to write in file the different states
import scipy.linalg as lg
import numpy as np
from .ldos import spatial_dos as spatial_wave
from .ldos import write_ldos as write_wave

def states0d(h,ewindow=[-.5,.5],signed=False,prefix=""):
  """Write in files the different states"""
  if not h.dimensionality==0: raise
  (evals,evecs) = lg.eigh(h.intra)
  evecs = evecs.transpose() # transpose list
  for i in range(len(evals)):
    if ewindow[0]<evals[i]<ewindow[1]: # if in the interval
      print("Printing state",evals[i])
      den = np.abs(evecs[i])**2
      den = spatial_wave(h,den) # resum if other degrees of freedom
      name=prefix+"WAVE_"+str(i)+"_energy_"+str(evals[i])+"_.OUT"
      write_wave(h.geometry.x,h.geometry.y,den,output_file=name)
      if signed: # if you want to write signed waves
        v = evecs[i]
        ii = np.sum(v.imag)
        if ii<0.00001:
          den = spatial_wave(h,v.real) # resum if other degrees of freedom   
          name="WAVE_SIGNED_"+str(i)+"_energy_"+str(evals[i])+".OUT"
          write_wave(h.geometry.x,h.geometry.y,den,output_file=name)
        else:
          print("Warning, non real function")




def states2d(h,ewindow=[-.5,.5],signed=False,prefix="",k=[0.,0.,0.],nrep=3):
  """Write in files the different states in 2d"""
  if not h.dimensionality==2: raise
  hk = h.get_hk_gen() # get generator
  m = hk(k) # stre this matrix
  (evals,evecs) = lg.eigh(m)
  evecs = evecs.transpose() # transpose list
  densum = h.geometry.x*0.0 # storage
  for i in range(len(evals)):
    if ewindow[0]<evals[i]<ewindow[1]: # if in the interval
      print("Printing state",evals[i])
      den = np.abs(evecs[i])**2
      den = spatial_wave(h,den) # resum if other degrees of freedom
      densum += den # store density
      name=prefix+"WAVE_"+str(i)+"_energy_"+str(evals[i])+"_.OUT"
      go = h.geometry.copy() # copy geometry
      go = go.supercell(nrep) # create supercell
      write_wave(go.x,go.y,den.tolist()*(nrep**2),output_file=name,z=go.z) 
      if signed: # if you want to write signed waves
        v = evecs[i]
        ii = np.sum(np.abs(v.imag))
        if ii<0.00001:
          den = spatial_wave(h,v.real) # resum if other degrees of freedom   
          name="WAVE_SIGNED_"+str(i)+"_energy_"+str(evals[i])+".OUT"
          go = h.geometry.copy() # copy geometry
          go = go.supercell(nrep) # create supercell
          write_wave(go.x,go.y,den.tolist()*(nrep**2),output_file=name,z=go.z) 
        else:
          print("Warning, non real function")
  go = h.geometry.copy() # copy geometry
  go = go.supercell(nrep) # create supercell
  write_wave(go.x,go.y,densum.tolist()*(nrep**2),output_file="WAVESUM.OUT",z=go.z) 






def multi_states(h,ewindow=[-0.5,0.5],signed=False,
          ks=None):
  """Calculate the bands and states in an energy window,
  and write them in a folder"""
  import os
  if ks is None:
    from . import klist
    ks = klist.default(h.geometry) # generate default klist
  os.system("rm -rf MULTISTATES_KPATH") # delete folder
  os.system("mkdir MULTISTATES_KPATH") # create folder
  os.chdir("MULTISTATES_KPATH") # go to the folder
  ik = 0
  for k in ks: # loop over kpoints
    print("MULTISTATES for",k)
    states2d(h,ewindow=ewindow,signed=signed,k=k,prefix="K_"+str(ik)+"_") 
    ik += 1
  os.chdir("..") # go to the folder







from numba import jit
import numpy as np


def crop_matrix(m,store):
    """Retain just some entries of a matrix"""
    if len(store)!=m.shape[0]: raise
    store = np.array(store,dtype=int) # transform to an integer
    n = np.sum(store) # number of entries of the matrix
    mout = np.zeros((n,n),dtype=np.complex)
    out = crop_matrix_jit(m,store,mout)
#    print(np.sum(np.abs(out))) 
    return out

@jit(nopython=True)
def crop_matrix_jit(m,store,mout):
    """Just retain the right elements"""
    ii = 0
    jj = 0
    n0 = m.shape[0] # input dimension
    n = mout.shape[0] # output dimension
    ii = 0
    for i in range(n0):
        if not store[i]==1: continue # skip iteration
        jj = 0 # initialize
        for j in range(n0):
            if store[j]==1: # store this one
                mout[ii,jj] = m[i,j]
                jj += 1
        ii += 1
    return mout








from __future__ import print_function
import numpy as np



# this library contains some checks for the hamiltonians


def electron_hole_symmetry(h,tol=1e-5,ntries=10):
   """Check that a system has electron-hole symmetry"""
   if not h.has_eh: return # skip if there is no e-h sector
   from superconductivity import eh_operator
   ehop = eh_operator(h.intra) # get the operator
   def ehsym(m1,m2):
     """Test eh symmetry"""
     diff = m1 + ehop(m2)
     diff = np.sum(np.abs(diff))
     if diff>tol: raise
   ehsym(h.intra,h.intra) # do it for the intra matrix
   if h.dimensionality>0:
     for i in range(ntries):
       hkgen = h.get_hk_gen() # generator
       k = np.random.random(h.dimensionality) # kpoint
       ehsym(hkgen(k),hkgen(-k))
       print("Checking k-point",k)
   print("Electron-hole symmetry is ok")
   





import numpy as np
import scipy.linalg as lg

def kfun_map(h,nk=50,
               nsuper=1,reciprocal=True,
               operator=None,k0=[0.,0.]):
    """ Calculate a reciprocal space map"""
    if operator is None: raise
    if h.dimensionality!=2: raise  # continue if two dimensional
    hk_gen = h.get_hk_gen() # gets the function to generate h(k)
    kxs = np.linspace(-nsuper,nsuper,nk)+k0[0]  # generate kx
    kys = np.linspace(-nsuper,nsuper,nk)+k0[1]  # generate ky
    kdos = [] # empty list
    kxout = []
    kyout = []
    if reciprocal: R = h.geometry.get_k2K() # get matrix
    else:  R = np.array(np.identity(3)) # get identity
    out = [] # empty list
    kx = []
    ky = []
    for x in kxs:
        for y in kxs:
            print("Doing",x,y)
            r = np.array([x,y,0.]) # real space vectors
            k = np.array(R)@r # change of basis
            hk = hk_gen(k) # get hamiltonian
            o = operator(hk) # make an operation at this kpoint
            out.append(o) # store
            kx.append(x) # store
            ky.append(y) # store
    return kx,ky,out # return result


def conduction_texture(h,n=2,**kwargs):
    """Compute the spin texture in the conduction band"""
    def fun(hk,O=None):
        """Function to call"""
        es,ws = lg.eigh(hk) # diagonalize
        ws = ws.T # transpose
        ws = ws[es>0.] # positive energies
        es = es[es>0.] # positive energies
        m = np.zeros((n,n),dtype=np.complex) # create matrix
        for i in range(n):
            wi = np.conjugate(O@ws[i])
            for j in range(n):
                m[i,j] = ws[j].dot(wi) # compute
        return m # return matrix
    fmx = lambda hk: fun(hk,O=h.get_operator("sx").get_matrix())
    fmy = lambda hk: fun(hk,O=h.get_operator("sy").get_matrix())
    kx,ky,mx = kfun_map(h,operator=fmx,**kwargs) # compute the mx matrix
    kx,ky,my = kfun_map(h,operator=fmy,**kwargs) # compute the my matrix
    # compute trace and determinant
    dmx = [lg.det(mxi).real for mxi in mx]
    dmy = [lg.det(myi).real for myi in my]
    tmx = [np.trace(mxi).real for mxi in mx]
    tmy = [np.trace(myi).real for myi in my]
    # now save the data
    np.savetxt("TRACE_TEXTURE.OUT",np.array([kx,ky,tmx,tmy]).T)
    np.savetxt("DET_TEXTURE.OUT",np.array([kx,ky,dmx,dmy]).T)







from __future__ import print_function
from __future__ import division
from .helptk import get_docstring
import numpy as np
from scipy.sparse import csc_matrix,bmat,csr_matrix
import scipy.linalg as lg
import scipy.sparse.linalg as slg
import numpy as np
from . import ldos
from . import operators
from . import inout
from . import superconductivity
from . import kanemele 
from . import magnetism
from . import checkclass
from . import extract
from . import multicell
from . import spectrum
from . import kekule
from . import algebra
from . import groundstate
from . import rotate_spin
from . import topology
from . import ldos
from . import increase_hilbert
from .meanfield import Vinteraction
from .sctk import dvector
from .algebratk import hamiltonianalgebra
from .bandstructure import get_bands_nd

from scipy.sparse import coo_matrix,bmat,csc_matrix
from .rotate_spin import sx,sy,sz
from .increase_hilbert import get_spinless2full,get_spinful2full
from . import tails
from scipy.sparse import diags as sparse_diag
import pickle
from .htk import mode as hamiltonianmode
from .htk import symmetry

#import data

from .limits import densedimension as maxmatrix
optimal = False

class Hamiltonian():
    """ Class for a hamiltonian """
    def get_tails(self,discard=None):
      """Write the tails of the wavefunctions"""
      if self.dimensionality!=0: raise
      else: return tails.matrix_tails(self.intra,discard=discard)
    def __add__(self,h):  return hamiltonianalgebra.add(self,h)
    def __rmul__(self,h):  return hamiltonianalgebra.rmul(self,h)
    def __mul__(self,h):  return hamiltonianalgebra.mul(self,h)
    def __neg__(self):  return (-1)*self
    def __sub__(self,a):  return self + (-a)
    def spinless2full(self,m,**kwargs):
        """Transform a spinless matrix in its full form"""
        return get_spinless2full(self,**kwargs)(m) # return
    def spinful2full(self,m):
        """Transform a spinless matrix in its full form"""
        return get_spinful2full(self)(m) # return
    def kchain(self,k=0.):
        return kchain(self,k)
    def get_fermi_surface(self,**kwargs):
        return spectrum.fermi_surface(self,**kwargs)
    def get_multi_fermi_surface(self,**kwargs):
        return spectrum.multi_fermi_surface(self,**kwargs)
    def get_eigenvectors(self,**kwargs):
        from .htk.eigenvectors import get_eigenvectors
        return get_eigenvectors(self,**kwargs)
    def get_gf(self,**kwargs):
        from .htk.green import get_gf
        return get_gf(self,**kwargs)
    def modify_hamiltonian_matrices(self,f,**kwargs):
        """Modify all the matrices of a Hamiltonian"""
        modify_hamiltonian_matrices(self,f,**kwargs)
    def add_strain(self,fs,**kwargs):
        from .strain import add_strain
        add_strain(self,fs,**kwargs) 
    def remove_sites(self,store):
        from . import sculpt
        self.geometry = sculpt.remove_sites(self.geometry,store)
        from .algebratk.matrixcrop import crop_matrix
        if self.has_spin: raise
        f = lambda m: crop_matrix(m,store)
        self.modify_hamiltonian_matrices(f) # modify all the matrices
    def get_filling(self,**kwargs):
        """Get the filling of a Hamiltonian at this energy"""
        return spectrum.get_filling(self,**kwargs) # eigenvalues
    def project_interactions(self,**kwargs):
        """Project interactions"""
        from .interactions.vijkl import Vijkl
        return Vijkl(self,**kwargs)
    def reduce(self):
        return hamiltonianmode.reduce_hamiltonian(self)

    def full2profile(self,x,**kwargs):
        """Transform a 1D array in the full space to the spatial basis"""
        from .htk.matrixcomponent import full2profile
        return full2profile(self,x,**kwargs)

    def get_chern(h,**kwargs):
        return topology.chern(h,**kwargs)

    def get_chi(self,**kwargs):
        from . import chi
        return chi.chiAB_trace(self,**kwargs)

    def get_hopping_dict(self):
        """Return the dictionary with the hoppings"""
        return multicell.get_hopping_dict(self)

    def get_multihopping(self):
        out = multicell.get_hopping_dict(self)
        return multicell.MultiHopping(out) # return the object

    def set_multihopping(self,mh):
        """Set a multihopping as the Hamiltonian"""
        multicell.set_multihopping(self,mh)
    def get_berry_curvature(self,**kwargs):
        return topology.write_berry(self,**kwargs)
    @get_docstring(spectrum.set_filling)
    def set_filling(self,filling,**kwargs):
        spectrum.set_filling(self,filling=filling,**kwargs)

    def __init__(self,geometry=None):
      self.data = dict() # empty dictionary with various data
      self.has_spin = True # has spin degree of freedom
      self.prefix = "" # a string used a prefix for different files
      self.path = "" # a path used for different files
      self.has_eh = False # has electron hole pairs
      self.get_eh_sector = None # no function for getting electrons
      self.fermi_energy = 0.0 # fermi energy at zero
      self.dimensionality = 0 # dimensionality of the Hamiltonian
      self.temperature = 0.0 # temperature of the Hamiltonian
      self.is_sparse = False
      self.is_multicell = False # for hamiltonians with hoppings to several neighbors
      self.hopping_dict = {} # hopping dictonary
      self.has_hopping_dict = False # has hopping dictonary
      self.non_hermitian = False # non hermitian Hamiltonian
      if not geometry is None:
  # dimensionality of the system
        self.dimensionality = geometry.dimensionality 
        self.geometry = geometry # add geometry object
        self.num_orbitals = len(geometry.x)
    def get_hk_gen(self):
        """ Generate kdependent hamiltonian"""
        #if self.is_multicell:
        out = multicell.hk_gen(self) # for multicell
       # else: out = hk_gen(self) # for normal cells
        from .htk.canonicalphase import canonical_unitary
#        return canonical_unitary(self,out)
        return out

    def has_time_reversal_symmetry(self):
        """Check if a Hamiltonian has time reversal symmetry"""
        from .htk import symmetry
        return symmetry.has_time_reversal_symmetry(self)

    def get_qpi(self,**kwargs):
        from .chitk import qpi
        return qpi.get_qpi(self,**kwargs)
    def get_rkky(self,**kwargs):
        #from .chitk import magneticresponse
        from . import rkky
        return rkky.rkky(self,**kwargs)
    @get_docstring(ldos.get_ldos)
    def get_ldos(self,**kwargs):
        return ldos.get_ldos(self,**kwargs)
    def get_gk_gen(self,delta=0.05,operator=None,canonical_phase=False):
      """Return the Green function generator"""
      hkgen = self.get_hk_gen() # Hamiltonian generator
      def f(k=[0.,0.,0.],e=0.0):
          hk = hkgen(k) # get matrix
          if canonical_phase: # use a Bloch phase in all the sites
            frac_r = self.geometry.frac_r # fractional coordinates
            # start in zero
            U = np.diag([self.geometry.bloch_phase(k,r) for r in frac_r])
            U = np.matrix(U) # this is without .H
            # increase the space if necessary
            U = self.spinless2full(U,is_hamiltonian=False) 
            Ud = algebra.dagger(U) # dagger
            hk = Ud@hk@U
          if operator is not None: 
              hk = algebra.dagger(operator)@hk@operator # project
          out = algebra.inv(np.identity(hk.shape[0])*(e+1j*delta) - hk)
          return out
      return f
    def to_canonical_gauge(self,m,k):
        """Return a matrix in the canonical gauge"""
        from . import gauge
        return gauge.to_canonical_gauge(self,m,k) # return the matrix
    def print_hamiltonian(self):
        """Print hamiltonian on screen"""
        print_hamiltonian(self)
    def check_mode(self,n):
        """Verify the type of Hamiltonian"""
        return hamiltonianmode.check_mode(self,n)
    def diagonalize(self,nkpoints=100):
      """Return eigenvalues"""
      return diagonalize(self,nkpoints=nkpoints)
    def get_fermi4filling(self,filling,**kwargs):
        """Return the fermi energy for a certain filling"""
        return spectrum.get_fermi4filling(self,filling,**kwargs)
    def get_dos(self,**kwargs):
        from . import dos
        return dos.get_dos(self,**kwargs)
    def get_density(self,**kwargs):
        from .ldostk import atomicmultildos
        return atomicmultildos.get_density(self,**kwargs)
    def get_bands(self,**kwargs):
      """ Returns a figure with teh bandstructure"""
      return get_bands_nd(self,**kwargs)
    def add_sublattice_imbalance(self,mass):
      """ Adds a sublattice imbalance """
      if self.geometry.has_sublattice and self.geometry.sublattice_number==2:
        add_sublattice_imbalance(self,mass)
      else: pass
    def add_antiferromagnetism(self,mass):
        """ Adds antiferromagnetic imbalanc """
        if self.geometry.has_sublattice:
            if self.geometry.sublattice_number==2:
                magnetism.add_antiferromagnetism(self,mass)
            elif self.geometry.sublattice_number>2:
                magnetism.add_frustrated_antiferromagnetism(self,mass)
            else: raise
        else: return 
    def turn_nambu(self):
        """Add electron hole degree of freedom"""
        self.get_eh_sector = get_eh_sector_odd_even # assign function
        superconductivity.turn_nambu(self)
    def add_swave(self,*args,**kwargs):
        """ Adds swave superconducting pairing"""
        superconductivity.add_swave_to_hamiltonian(self,*args,**kwargs)
    def setup_nambu_spinor(self): self.add_swave(0.0)
    def get_anomalous_hamiltonian(self):
        """Return a Hamiltonian only with the anomalous part"""
        return superconductivity.get_anomalous_hamiltonian(self)
    def add_pairing(self,**kwargs):
        """ Add a general pairing matrix, uu,dd,ud"""
        superconductivity.add_pairing_to_hamiltonian(self,**kwargs)
    def same_hamiltonian(self,*args,**kwargs):
        """Check if two hamiltonians are the same"""
        return hamiltonianmode.same_hamiltonian(self,*args,**kwargs)
    def get_supercell(self,nsuper,**kwargs):
      """ Creates a supercell of a one dimensional system"""
      if nsuper==1: return self
      if self.dimensionality==0: return self
      try: 
          nsuper[2] 
          ns = nsuper # array as input
      except:
          if self.dimensionality==1: ns = [nsuper,1,1]
          elif self.dimensionality==2: ns = [nsuper,nsuper,1]
          elif self.dimensionality==3: ns = [nsuper,nsuper,nsuper]
          else: raise
      return multicell.supercell_hamiltonian(self,nsuper=ns,**kwargs)
    def supercell(self,*args,**kwargs):
      return self.get_supercell(*args,**kwargs)
    def set_finite_system(self,periodic=True):
      """ Transforms the system into a finite system"""
      return set_finite_system(self,periodic=periodic) 
    def get_gap(self,**kwargs):
      """Returns the gap of the Hamiltonian"""
      from . import gap
      return gap.get_gap(self,**kwargs) # return the gap
    def save(self,output_file="hamiltonian.pkl"):
      """ Write the hamiltonian in a file"""
      inout.save(self,output_file) # write in a file
    write = save # just in case
    def read(self,output_file="hamiltonian.pkl"):
      """ Read the Hamiltonian"""
      return load(output_file) # read Hamiltonian

    def load(self,**kwargs): return self.read(**kwargs)

    @get_docstring(spectrum.total_energy)
    def get_total_energy(self,**kwargs):
      return spectrum.total_energy(self,**kwargs)

    def total_energy(self,**kwargs): return self.get_total_energy(**kwargs)

    def add_zeeman(self,zeeman):
        """Adds zeeman to the matrix """
        self.turn_spinful()
        from .magnetism import add_zeeman
        add_zeeman(self,zeeman=zeeman)
    def add_magnetism(self,m):
        """Adds magnetism, new version of zeeman"""
        self.turn_spinful()
        from .magnetism import add_magnetism
        add_magnetism(self,m)
    def add_exchange(self,m): self.add_magnetism(m)
    def turn_spinful(self,**kwargs):
        from .htk.mode import turn_spinful
        turn_spinful(self,**kwargs)
    def remove_spin(self):
      """Removes spin degree of freedom"""
      if self.check_mode("spinless"): return # do nothing
      elif self.check_mode("spinful"):
          def f(m): return des_spin(m,component=0)
          self.modify_hamiltonian_matrices(f) # modify the matrices
          self.has_spin = False # set to spinless
      else: raise
    def remove_nambu(self):
      if self.check_mode("spinful_nambu"): 
          def f(m):
              return superconductivity.get_eh_sector(m,i=0,j=0)
          self.modify_hamiltonian_matrices(f) # modify the matrices
          self.has_eh = False # set to normal
      elif self.check_mode("spinful"): pass
      elif self.check_mode("spinless"): pass
      else: raise
    def add_onsite(self,fermi):
      """ Move the Fermi energy of the system"""
      shift_fermi(self,fermi)
    def get_topological_invariant(self,**kwargs):
        """Return a topological invariant"""
        if self.dimensionality==0: pass
        elif self.dimensionality==1: pass
        elif self.dimensionality==2: 
            if self.has_time_reversal_symmetry():
                return topology.z2_invariant(self,**kwargs)
            else:
                print("Computing Chern")
                return topology.chern(self,**kwargs)
        else: raise
    def shift_fermi(self,fermi): self.add_onsite(fermi)  
    def first_neighbors(self):
      """ Create first neighbor hopping"""
      if 0<=self.dimensionality<3:
        first_neighborsnd(self)
      elif self.dimensionality == 3:
        from .multicell import first_neighbors as fnm
        fnm(self)
      else: raise
    def add_hopping_matrix(self,fm):
        """
        Add a certain hopping matrix to the Hamiltonian
        """
        if not self.is_multicell: 
            self.turn_multicell()
            #raise # this may not work for multicell
        h = self.geometry.get_hamiltonian(has_spin=self.has_spin,
                is_multicell=self.is_multicell,
                mgenerator=fm) # generate a new Hamiltonian
        self.add_hamiltonian(h) # add this contribution
    def add_hamiltonian(self,h):
        """
        Add the hoppings of another Hamiltonian
        """
        if not self.is_multicell: # not implemented
            self.turn_multicell()
        hd = h.get_dict() # get the dictionary
        self.intra = self.intra + hd[(0,0,0)] # add the matrix
        for i in range(len(self.hopping)):
            d = tuple(self.hopping[i].dir)
            if d in hd:
              self.hopping[i].m = self.hopping[i].m + hd[d]
    def get_dict(self):
        """
        Return the dictionary that yields the hoppings
        """
        if not self.is_multicell: # not implemented
            self = self.get_multicell()
        hop = dict()
        hop[(0,0,0)] = self.intra
        for t in self.hopping: hop[tuple(np.array(t.dir))] = t.m
        for t in self.hopping: hop[tuple(-np.array(t.dir))] = np.conjugate(t.m).T
        return hop # return dictionary

    @get_docstring(ldos.multi_ldos)
    def get_multildos(self,**kwargs):
        return ldos.multi_ldos(self,**kwargs)
    def get_multihopping(self):
        """Return a multihopping object"""
        from .multihopping import MultiHopping
        return MultiHopping(self.get_dict())

    @get_docstring(Vinteraction)
    def get_mean_field_hamiltonian(self,**kwargs):
        return Vinteraction(self,**kwargs).hamiltonian
    def copy(self):
        """
        Return a copy of the hamiltonian
        """
        from copy import deepcopy
        return deepcopy(self)
    def check(self,**kwargs):
        """
        Check if the Hamiltonian is OK
        """
        from . import check
        check.check_hamiltonian(self,**kwargs) # check the Hamiltonian
    def enforce_eh(self):
        """Enforce electron-hole symmetry in the Hamiltonian"""
        self.turn_multicell() # turn to multicell mode
        from superconductivity import eh_operator
        f = eh_operator(self.intra) # electron hole operator
        raise # not implemented
    def turn_sparse(self):
        """
        Transforms the hamiltonian into a sparse hamiltonian
        """
        from scipy.sparse import csc_matrix
        def f(m):
            return csc_matrix(m)
        self.modify_hamiltonian_matrices(f) # modify the matrices
        self.is_sparse = True # sparse flag to true
    def turn_dense(self):
      """ Transforms the hamiltonian into a sparse hamiltonian"""
      def f(m):
          return algebra.todense(m)
      self.modify_hamiltonian_matrices(f) # modify the matrices
      self.is_sparse = False # sparse flag to true
    def add_rashba(self,c):
      """Adds Rashba coupling"""
      from . import rashba
      rashba.add_rashba(self,c)
    def add_soc(self,t,**kwargs):
      self.add_kane_mele(t,**kwargs)
    def add_kane_mele(self,t,**kwargs):
      """ Adds a Kane-Mele SOC term"""  
      kanemele.add_kane_mele(self,t,**kwargs) # return kane-mele SOC
    def add_haldane(self,t):
      """ Adds a Haldane term"""  
      kanemele.add_haldane(self,t) # return Haldane SOC
    def add_kekule(self,t):
        """
        Add Kekule coupling
        """
        if self.dimensionality==0: # zero dimensional
          m = kekule.kekule_matrix(self.geometry.r,t=t)
          self.intra = self.intra + self.spinless2full(m)
        else: # workaround for higher dimensionality
          r = self.geometry.multireplicas(2) # get many replicas
          fm = kekule.kekule_function(r,t=t)
          self.add_hopping_matrix(fm) # add the Kekule hopping
    def add_chiral_kekule(self,**kwargs):
        """
        Add a chiral kekule hopping
        """
        fun = kekule.chiral_kekule(self.geometry,**kwargs)
        self.add_kekule(fun)
  
    def add_modified_haldane(self,t):
        """
        Adds a Haldane term
        """  
        kanemele.add_modified_haldane(self,t) # return Haldane SOC
    def add_anti_kane_mele(self,t):
        """
        Adds an anti Kane-Mele term
        """  
        kanemele.add_anti_kane_mele(self,t) # return anti kane mele SOC
    def add_antihaldane(self,t): 
        """Add an anti-Haldane term"""
        self.add_modified_haldane(t) # second name
    def add_crystal_field(self,v,**kwargs):
        """Add a crystal field term to the Hamiltonian"""
        from . import crystalfield
        crystalfield.hartree(self,v=v,**kwargs) 
    def add_peierls(self,mag_field,**kwargs):
        """
        Add magnetic field
        """
        from .peierls import add_peierls
        add_peierls(self,mag_field=mag_field,**kwargs)
    def add_inplane_bfield(self,**kwargs):
        """Add in-plane magnetic field"""
        from .peierls import add_inplane_bfield
        add_inplane_bfield(self,**kwargs)
    def align_magnetism(self,vectors):
        """ Rotate the Hamiltonian to have magnetism in the z direction"""
        if self.has_eh: raise
        from .rotate_spin import align_magnetism as align
        self.intra = align(self.intra,vectors)
        self.inter = align(self.inter,vectors)
    def global_spin_rotation(self,**kwargs):
        """ Perform a global spin rotation """
        return rotate_spin.hamiltonian_spin_rotation(self,**kwargs)
    def generate_spin_spiral(self,**kwargs):
        """ Generate a spin spiral antsaz in the Hamiltonian """
        return rotate_spin.generate_spin_spiral(self,**kwargs)
    def get_magnetization(self,**kwargs):
        """ Return the magnetization """
        mx = self.extract(name="mx")
        my = self.extract(name="my")
        mz = self.extract(name="mz")
        return np.array([mx,my,mz]).T # return array
    def compute_vev(self,name="sz",**kwargs):
        """
        Compute a VEV of a spatially resolved operator
        """
        n = len(self.geometry.r) # number of sites
        ops = [operators.index(self,n=[i]) for i in range(n)]
        if name=="sx": op = operators.get_sx(self)
        elif name=="sy": op = operators.get_sy(self)
        elif name=="sz": op = operators.get_sz(self)
        elif name=="density": op = operators.index(self,n=range(n))
        else: raise
        ops = [o@op for o in ops] # define operators
        return spectrum.ev(self,operator=ops,**kwargs).real
    def get_1dh(self,k=0.0):
        """Return a 1d Hamiltonian"""
        if self.is_multicell: # not implemented
            self = self.get_no_multicell() # return the no multicell Hamiltonian
        if not self.dimensionality==2: raise # not implemented
        intra,inter = kchain(self,k) # generate intra and inter
        hout = self.copy() # copy the Hamiltonian
        hout.intra = intra # store
        hout.inter = inter # store
        hout.dimensionality = 1 # one dimensional
        hout.geometry.dimensionality = 1 # one dimensional
        return hout
    def get_multicell(self):
        """Return a multicell Hamiltonian"""
        return multicell.turn_multicell(self)
    def turn_multicell(self):
        """Conver to multicell Hamiltonian"""
        h = multicell.turn_multicell(self)
        self.is_multicell = True
        self.hopping = h.hopping # assign hopping
    def get_no_multicell(self):
        """Return a multicell Hamiltonian"""
        h1 = multicell.turn_no_multicell(self)
        h0 = h1.get_multicell() # turn multicell again
        diff = (self.get_multihopping() - h0.get_multihopping()).norm()
        if diff>1e-6: 
            print("Hamiltonain cannot be made no multicell")
            raise
        else: return h1 # return the Hamiltonian
    def clean(self):
        """Clean a Hamiltonian"""
        from .clean import clean_hamiltonian
        clean_hamiltonian(self)
    def get_operator(self,name,**kwargs):
        """Return a certain operator"""
        from . import operatorlist
        return operators.object2operator(operatorlist.get_operator(self,name,
            **kwargs))
    def extract(self,name): 
        """Extract something from the Hamiltonian"""
        return extract.extract_from_hamiltonian(self,name)
    @get_docstring(dvector.dvector_non_unitarity_map)
    def write_non_unitarity(self,**kwargs):
        dvector.dvector_non_unitarity_map(self,**kwargs)
    def write_magnetization(self,nrep=5):
        """Extract the magnetization and write it in a file"""
        mx = self.extract("mx")
        my = self.extract("my")
        mz = self.extract("mz")
        g = self.geometry
        g.write_profile(mx,name="MX.OUT",normal_order=True,nrep=nrep)
        g.write_profile(my,name="MY.OUT",normal_order=True,nrep=nrep)
        g.write_profile(mz,name="MZ.OUT",normal_order=True,nrep=nrep)
        # this is just a workaround
        m = np.genfromtxt("MX.OUT").transpose()
        (x,y,z,mx) = m[0],m[1],m[2],m[3]
        my = np.genfromtxt("MY.OUT").transpose()[3]
        mz = np.genfromtxt("MZ.OUT").transpose()[3]
        np.savetxt("MAGNETISM.OUT",np.array([x,y,z,mx,my,mz]).T)
        return np.array([mx,my,mz])
    def write_onsite(self,nrep=5,normal_order=False):
        """Extract onsite energy"""
        d = self.extract("density")
        d = d - np.mean(d)
        self.geometry.write_profile(d,name="ONSITE.OUT",
                normal_order=normal_order,nrep=nrep)
    def write_hopping(self,**kwargs):
        groundstate.hopping(self,**kwargs)
    def write_anomalous_hopping(self,**kwargs):
        groundstate.anomalous_hopping(self,**kwargs)
    def write_swave(self,**kwargs):
        """Write the swave pairing"""
        groundstate.swave(self,**kwargs)
    def get_ipr(self,**kwargs):
        """Return the IPR"""
        from . import ipr
        if self.dimensionality==0:
            return ipr.ipr(self.intra,**kwargs) 
        else: raise # not implemented
    @get_docstring(dvector.dvector_non_unitarity)
    def get_dvector_non_unitarity(self,**kwargs):
        return dvector.dvector_non_unitarity(self,**kwargs)
    def get_density_matrix(self,**kwargs):
        """Return the density matrix"""
        from . import densitymatrix
        return densitymatrix.full_dm(self,**kwargs)
    @get_docstring(superconductivity.average_hamiltonian_dvector)
    def get_average_dvector(self,**kwargs):
        return superconductivity.average_hamiltonian_dvector(self,**kwargs)


hamiltonian = Hamiltonian





def get_first_neighbors(r1,r2):
    """Gets the fist neighbors, input are arrays"""
    from . import neighbor
    pairs = neighbor.find_first_neighbor(r1,r2)
    return pairs





def create_fn_hopping(r1,r2):
  n=len(r1)
  mat=np.matrix([[0.0j for i in range(n)] for j in range(n)])
  pairs = get_first_neighbors(r1,r2) # get pairs of first neighbors
  for p in pairs: # loop over pairs
    mat[p[0],p[1]] = 1.0 
  return mat



# function to calculate the chirality between two vectors (vectorial productc)
def vec_chi(r1,r2):
  """Return clockwise or anticlockwise"""
  z=r1[0]*r2[1]-r1[1]*r2[0]
  zz = r1-r2
  zz = sum(zz*zz)
  if zz>0.01:
    if z>0.01: # clockwise
      return 1.0
    if z<-0.01: # anticlockwise
      return -1.0
  return 0.0




# routine to check if two atoms arein adistance d
def is_neigh(r1,r2,d,tol):
  r=r1-r2
  x=r[0]
  y=r[1]
  dt=abs(d*d-x*x-y*y)
  if dt<tol:
    return True
  return False




#################################3

def diagonalize(h,nkpoints=100):
  """ Diagonalice a hamiltonian """
  import scipy.linalg as lg
  # for one dimensional systems
  if h.dimensionality==1:  # one simensional system
    klist = np.arange(0.0,1.0,1.0/nkpoints)  # create array with the klist
    if h.geometry.shift_kspace:
      klist = np.arange(-0.5,0.5,1.0/nkpoints)  # create array with the klist
    intra = h.intra  # assign intraterm
    inter = h.inter  # assign interterm
    energies = [] # list with the energies
    for k in klist: # loop over kpoints
      bf = np.exp(1j*np.pi*2.*k)  # bloch factor for the intercell terms
      inter_k = inter*bf  # bloch hopping
      hk = intra + inter_k + inter_k.H # k dependent hamiltonian
      energies += [lg.eigvalsh(hk)] # get eigenvalues of the current hamiltonian
    energies = np.array(energies).transpose() # each kpoint in a line
    return (klist,energies) # return the klist and the energies
# for zero dimensional systems system
  elif h.dimensionality==0:  
    intra = h.intra  # assign intraterm
    energies = lg.eigvalsh(intra) # get eigenvalues of the current hamiltonian
    return (range(len(intra)),energies) # return indexes and energies
  else: raise




def diagonalize_hk(k):
  return algebra.eigh(hk(k))




def diagonalize_kpath(h,kpath):
  """Diagonalice in a certain path"""
  energies = [] # empty list with energies
  import scipy.linalg as lg
  ik = 0.
  iks = [] # empty list
  for k in kpath:
    f = h.get_hk_gen() # get Hk generator
    hk = f(k)  # k dependent hamiltonian
    es = (lg.eigvalsh(hk)).tolist() # get eigenvalues for current hamiltonian
    energies += es # append energies 
    iks += [ik for i in es]
    ik += 1.
  iks = np.array(iks)
  iks = iks/max(iks) # normalize path
  return (iks,energies)





def print_hamiltonian(h):
  """ Print the hamilotnian on screen """
  from scipy.sparse import coo_matrix as coo # import sparse matrix
  intra = coo(h.intra) # intracell
  inter = coo(h.inter) # intracell
  print("Intracell matrix")
  print(intra)
  print("Intercell matrix")
  print(inter)
  return


from .htk.cdw import add_sublattice_imbalance


def build_eh_nonh(hin,c1=None,c2=None):
  """Creates a electron hole matrix, from an input matrix, coupling couples
     electrons and holes
      - hin is the hamiltonian for electrons, which has the usual common form
      - coupling is the matrix which tells the coupling between electron
        on state i woth holes on state j, for exmaple, with swave pairing
        the non vanishing elments are (0,1),(2,3),(4,5) and so on..."""
  n = len(hin)  # dimension of input
  nn = 2*n  # dimension of output
  hout = np.matrix(np.zeros((nn,nn),dtype=complex))  # output hamiltonian
  for i in range(n):
    for j in range(n):
      hout[2*i,2*j] = hin[i,j]  # electron term
      hout[2*i+1,2*j+1] = -np.conjugate(hin[i,j])  # hole term
  if not c1 is None: # if there is coupling
    for i in range(n):
      for j in range(n):
        # couples electron in i with hole in j
        hout[2*i,2*j+1] = c1[i,j]  # electron hole term
  if not c2 is None: # if there is coupling
    for i in range(n):
      for j in range(n):
        # couples hole in i with electron in j
        hout[2*j+1,2*i] = np.conjugate(c2[i,j])  # hole electron term
  return hout 









def set_finite_system(hin,periodic=True):
  """ Transforms the hamiltonian into a finite system,
  removing the hoppings """
  from copy import deepcopy
  h = hin.copy() # copy Hamiltonian
  h.dimensionality = 0 # put dimensionality = 0
  if periodic: # periodic boundary conditions
    if h.dimensionality == 1:
      h.intra = h.intra + h.inter + h.inter.H 
    if h.dimensionality == 2:
      h.intra = h.intra +  h.tx + h.tx.H 
      h.intra = h.intra +  h.ty + h.ty.H 
      h.intra = h.intra +  h.txy + h.txy.H 
      h.intra = h.intra +  h.txmy + h.txmy.H 
  else: pass
  return h
  
# remove spin degree of freedom
des_spin = increase_hilbert.des_spin


def shift_fermi(h,fermi):
  """ Moves the fermi energy of the system, the new value is at zero"""
  r = h.geometry.r # positions
  n = len(r) # number of sites
  if checkclass.is_iterable(fermi): # iterable
    if len(fermi)==n: # same number of sites
      h.intra = h.intra + h.spinless2full(sparse_diag([fermi],[0]))
    else: raise
  else:
    rc = [i for i in range(n)]  # index
    datatmp = [] # data
    for i in range(n): # loop over positions 
      if callable(fermi): fshift = fermi(r[i]) # fermi shift
      else: fshift = fermi # assume it is a number
      datatmp.append(fshift) # append value
    m = csc_matrix((datatmp,(rc,rc)),shape=(n,n)) # matrix with the shift
    h.intra = h.intra + h.spinless2full(m) # Add matrix 
    return


import numbers

def is_number(s):
    return isinstance(x, numbers.Number)

def is_hermitic(m):
  mh = np.conjugate(m).T
  hh = m - mh
  for i in range(len(hh)):
    for j in range(len(hh)):
      if np.abs(hh[i,j]) > 1e-5:
        print("No hermitic element", i,j,m[i,j],m[j,i])
        return False
  return True
  





def first_neighborsnd(h):
  """ Gets a first neighbor hamiltonian"""
  r = h.geometry.r    # x coordinate 
  g = h.geometry
# first neighbors hopping, all the matrices
  a1, a2 = g.a1, g.a2
  def gett(r1,r2):
    """Return hopping given two sets of positions"""
    from . import neighbor
    pairs = neighbor.find_first_neighbor(r1,r2)
    if len(pairs)==0: rows,cols = [],[]
    else: rows,cols = np.array(pairs).T # transpose
    data = np.array([1. for c in cols])
    n = len(r1)
    m = csc_matrix((data,(rows,cols)),shape=(n,n),dtype=np.complex)
    m = h.spinless2full(m) # add spin degree of freedom if necessary
    if h.is_sparse: return m
    else: return m.todense()
  if h.dimensionality==0:
    h.intra = gett(r,r)
  elif h.dimensionality==1:
    h.intra = gett(r,r)
    h.inter = gett(r,r+a1)
  elif h.dimensionality==2:
    h.intra = gett(r,r)
    h.tx = gett(r,r+a1)
    h.ty = gett(r,r+a2)
    h.txy = gett(r,r+a1+a2)
    h.txmy = gett(r,r+a1-a2)
  else: raise



from .superconductivity import add_swave
from .superconductivity import build_eh
nambu_nonh = build_eh_nonh



from .bandstructure import lowest_bands



def hk_gen(h):
  """ Returns a function that generates a k dependent hamiltonian"""
  if h.dimensionality == 0: return lambda x: h.intra
  elif h.dimensionality == 1: 
    def hk(k):
      """k dependent hamiltonian, k goes from 0 to 1"""
      try: kp = k[0]
      except: kp = k
      tk = h.inter * h.geometry.bloch_phase([1.],kp) # get the bloch phase
      ho = h.intra + tk + algebra.dagger(tk) # hamiltonian
      return ho
    return hk  # return the function
  elif h.dimensionality == 2: 
    def hk(k):
      """k dependent hamiltonian, k goes from (0,0) to (1,1)"""
      if len(k)==3:
        k = np.array([k[0],k[1]]) # redefine for 2d
      k = np.array(k)
      ux = np.array([1.,0.])
      uy = np.array([0.,1.])
      ptk = [[h.tx,ux],[h.ty,uy],[h.txy,ux+uy],[h.txmy,ux-uy]] 
      ho = (h.intra).copy() # intraterm
      for p in ptk: # loop over hoppings
#        tk = p[0]*np.exp(1j*np.pi*2.*(p[1].dot(k)))  # add bloch hopping
        tk = p[0]*h.geometry.bloch_phase(p[1],k)  # add bloch hopping
        ho = ho + tk + algebra.dagger(tk)  # add bloch hopping
      return ho
    return hk
  else: raise



from .neighbor import parametric_hopping
from .neighbor import parametric_hopping_spinful
from .neighbor import generate_parametric_hopping




from .superconductivity import get_nambu_tauz
from .superconductivity import project_electrons
from .superconductivity import project_holes
from .superconductivity import get_eh_sector_odd_even




def kchain(h,k):
  """ Return the kchain Hamiltonian """
  if h.dimensionality != 2: raise
  if h.is_multicell: h = h.get_no_multicell() # redefine
  tky = h.ty*np.exp(1j*np.pi*2.*k)
  tkxy = h.txy*np.exp(1j*np.pi*2.*k)
  tkxmy = h.txmy*np.exp(-1j*np.pi*2.*k)  # notice the minus sign !!!!
  # chain in the x direction
  ons = h.intra + tky + np.conjugate(tky).T  # intra of k dependent chain
  hop = h.tx + tkxy + tkxmy  # hopping of k-dependent chain
  return (ons,hop)



# import the function written in the library
from .kanemele import generalized_kane_mele
from .htk.modify import modify_hamiltonian_matrices

from . import inout

def load(input_file="hamiltonian.pkl"):  return inout.load(input_file)


def print_hopping(h):
    """Print all the hoppings in a user friendly way"""
    from pandas import DataFrame
    def pprint(m): 
        if np.max(np.abs(m.real))>0.00001: print(DataFrame(m.real))
        if np.max(np.abs(m.imag))>0.00001: print(DataFrame(m.imag*1j))
        print("\n")
    print("Onsite")
    pprint(h.intra)
    if h.dimensionality==0: return
    h = h.get_multicell()
    for t in h.hopping:
        print("Hopping",t.dir)
        pprint(t.m)




from .htk.dummy import generate_dummy_hamiltonian
generate_hamiltonian_from_dict = generate_dummy_hamiltonian





import numpy as np
from . import operators

def to_canonical_gauge(self,m,k):
    """Return a matrix in the canonical gauge form"""
    frac_r = self.geometry.frac_r # fractional coordinates
    # start in zero
    U = np.diag([self.geometry.bloch_phase(k,r) for r in frac_r])
    U = np.array(U) # this is without .H
    U = self.spinless2full(U) # increase the space if necessary
    Ud = np.conjugate(U.T) # dagger
    out = Ud@m@U # redefine matrix
    return out


def canonical_gauge_transformation(self,k):
    """Return a matrix in the canonical gauge form"""
    frac_r = self.geometry.frac_r # fractional coordinates
    # start in zero
    U = np.diag([self.geometry.bloch_phase(k,r) for r in frac_r])
    U = np.array(U) # this is without .H
    U = self.spinless2full(U) # increase the space if necessary
    return U

def Operator2canonical_gauge(h,op):
    """transform an operator into canonical gauge, assuming that
    it is linear"""
    U = operators.Operator(lambda v,k=None: canonical_gauge_transformation(h,k)@v)
    Ud = operators.Operator(lambda v,k=None: canonical_gauge_transformation(h,k).T.conjugate()@v)
    return Ud*op*U


def hamiltonian_gauge_transformation(h,phis):
    ho = h.copy()
    phis = np.array(phis)
    if len(phis)!=len(h.geometry.r): raise
    U = np.diag(np.exp(1j*np.pi*2*phis)) # create the 
    U = ho.spinless2full(U) # increase the space if necessary
    Uh = np.conjugate(U.T) # transpose
    ho.modify_hamiltonian_matrices(lambda m: U@m@Uh)
    return ho






from __future__ import print_function
from . import geometry
from copy import deepcopy
from . import algebra
from numba import jit
import numpy as np
try: 
  import clean_geometryf90
  use_fortran = True
except: use_fortran = False


def remove(g,l):
  """ Remove certain atoms from the geometry"""
  nr = len(l) # number of removed atoms
  go = g.copy() # copy the geometry
  xo = [] # copy the list
  yo = [] # copy the list
  zo = [] # copy the list
  for i in range(len(g.x)):
    if not i in l:
      xo.append(g.x[i])
      yo.append(g.y[i])
      zo.append(g.z[i])
  go.x = np.array(xo)
  go.y = np.array(yo)
  go.z = np.array(zo)
  go.xyz2r() # update the revectors
  ##### if has sublattice ####
  if g.has_sublattice: # if has sublattice, keep the indexes
    ab = [] # initialize
    for i in range(len(g.x)):
      if not i in l:
        ab.append(g.sublattice[i]) # keep the index
    go.sublattice = ab # store the keeped atoms
  return go

def intersec(g,f):
  """ Intersec coordinates with a certain function which yields True or False,
  output is resultant geometry """
  store = np.array([f(ir) for ir in g.r],dtype=int) # store
  return remove_sites(g,store) # remove those sites


def remove_sites(g,store):
  """ Intersec coordinates with a certain function which yields True or False,
  output is resultant geometry """
  gout = g.copy() # copy the geometry
  store = np.array(store,dtype=int) # store
  gout.r = g.r[store==1]
  gout.r2xyz() # update r
  if gout.has_sublattice: # if has sublattice, keep the indexes
    gout.sublattice = g.sublattice[store==1]
  return gout


def intersected_indexes(g,f):
  """Return the indexes of the atoms located in this function"""
  iis = []
  for (ii,ix,iy,iz) in zip(range(len(g.x)),g.x,g.y,g.z): # loop over positions
    if f([ix,iy,iz]): # if the function yields true
      iis.append(ii)
  return iis











def circle(r=1.0,out=False):
  """ Returns a function which encondes a circle"""
  if out:  # if true is inside
    def f(x,y):
      if r*r > x*x + y*y:
        return True
      else:
        return False  
  else: # if true is outside
    def f(x,y):
      if r*r < x*x + y*y:
        return True
      else:
        return False  
  return f # return the function



def rotate(g,angle):
  """ Rotates a geometry"""
  if np.abs(angle)<0.0000001: 
#    print("No rotation performed")
    return g
  phi = angle
  go = g.copy()
  # modify x and y, z is the same
  c,s = np.cos(phi), np.sin(phi)  # sin and cos of the anggle
  go.x = c*g.x + s*g.y    # x coordinate 
  go.y = -s*g.x + c*g.y    # y coordinate
  go.xyz2r() # update r
  if go.dimensionality==2:  # two dimensional
    x,y,z = go.a1
    go.a1 = np.array([c*x + s*y,-s*x + c*y,z])
    x,y,z = go.a2
    go.a2 = np.array([c*x + s*y,-s*x + c*y,z])
  elif go.dimensionality==1: 
    x,y,z = go.a1
    go.a1 = np.array([c*x + s*y,-s*x + c*y,z])
  elif go.dimensionality==0: pass
  else: raise # 
  go.get_fractional() # get fractional coordinates 
  return go


def center(g,angle):
  """Center a geometry"""
  g.x = g.x -sum(g.x)/len(g.x)
  g.y = g.y -sum(g.y)/len(g.y)
 


def remove_unibonded(g,d=1.0,tol=0.01,use_fortran=use_fortran,iterative=False):
  """Removes from the geometry atoms with only one bond"""
  sb = []
  if g.dimensionality>0: use_fortran = False # only for 0d
  if use_fortran: # use fortran routine
# array with true in right atoms
    retain = clean_geometryf90.clean_geometry(g.r.transpose(),2) 
    sb = [] # remove this atoms
    for i in range(len(retain)):
      if not retain[i]: sb.append(i) # remove
    gout = remove(g,sb) # remove those atoms
  else: # use python routine
    for i in range(len(g.r)): 
      r1 = g.r[i] # first position
      nb = 0 # initialize
      for direc in g.neighbor_directions(): # loop over directions
        for r2 in g.replicas(d=direc): # loop over replicas
          dr = r1-r2
          if d-tol < dr.dot(dr) < d+tol: # if first neighbor
            nb += 1 # increase counter
      if nb<2:
        sb.append(i+0) # add to the list
    gout = remove(g,sb) # remove those atoms
  if iterative: # ensure that it hs the same number of atoms by calling again
    if len(g.x) != len(gout.x): # call again
      print("Iterative cleaning")
      return remove_unibonded(gout,d=d,tol=tol,
         use_fortran=use_fortran,iterative=iterative)
  return gout # return the geometry



def remove_central(g,n):
  """Removes n atoms from the center of the crystal"""
  rr = [r.dot(r) for r in g.r] # norm of the distances
  inds = range(len(rr)) # indexes
  sort_inds = [x for (y,x) in sorted(zip(rr,inds))] # indexes sorted by distance
  rind = [sort_inds[i] for i in range(n)]
  return remove(g,rind)  # return geometry with removed 


def get_closest(g,n=1,r0=[0.,0.,0.]):
  """Gets n atoms from the center of the crystal"""
  r0 = np.array(r0)
  rr = [(r-r0).dot(r-r0) for r in g.r] # norm of the distances
  inds = range(len(rr)) # indexes
  sort_inds = [x for (y,x) in sorted(zip(rr,inds))] # indexes sorted by distance
  rind = [sort_inds[i] for i in range(n)]
  return rind # return the indexes


def get_central(g,n=1):
  """Get the index of the central atom"""
  return get_closest(g,n=n,r0=[0.,0.,0.])

def shift(g,r=np.array([0.,0.,0.])):
  """Shift the geometry by a certain vector"""
  g.r = np.array([ri-r for ri in g.r])
  g.r2xyz() # update


def get_angle(v1,v2):
  """Get the angle between two vectors"""
  v3 = v1/np.sqrt(v1.dot(v1)) # normalize
  v4 = v2/np.sqrt(v2.dot(v2)) # normalize
  alpha = np.arccos(v3.dot(v4))
  return alpha





def get_furthest(g,n=1,angle=0.,tol=5.):
  """Gets n atoms in a certain direction"""
  rs = [] # norm of the distances
  inds = [] # norm of the distances
  for ir in range(len(g.r)): # store only vectors with a certain angle
    r = g.r[ir]
    a = np.arctan2(r[1],r[0])/np.pi*180.
    if (np.abs(angle-a)%360)<tol: # if direction is right
      rs.append(r) # store vector
      inds.append(ir) # indexes
  rr = [-r.dot(r) for r in rs] # norm of the distances
  sort_inds = [x for (y,x) in sorted(zip(rr,inds))] # indexes sorted by distance
  rind = [sort_inds[i] for i in range(n)]
  return rind # return the indexes





def rotate_a2b(g,a,b):
  """ Rotates the geometry making a original vector a pointing along b"""
  da = a.dot(a)
  da = a/np.sqrt(da) # unit vector
  db = b.dot(b)
  db = b/np.sqrt(db) # unit vector
  # define two complex numbers
  za = da[0] + da[1]*1j
  zb = db[0] + db[1]*1j
  angle = np.angle(za/zb) # angle in the complex plane
  return rotate(g,angle)


def build_island(gin,n=5,angle=20,nedges=6,clear=True):
  """ Build an island starting from a 2d geometry"""
  nf = float(n)   # get the desired size, in float
  if gin.dimensionality!=2: raise 
  g = gin.copy()
  g = g.supercell(8*n)   # create supercell
  g.set_finite() # set as finite system
  g.center() # center the geometry
  # now scuplt the geometry
  g = rotate(g,angle*2.*np.pi/360) # initial rotation
  def f(x,y): return x>-nf*(np.cos(np.pi/3)+1.)  # function to use as cut
  for i in range(nedges): # loop over rotations, 60 degrees
    g = intersec(g,f) # retain certain atoms
    g = rotate(g,2.*np.pi/nedges) # rotate 60 degrees
  if clear:  g = remove_unibonded(g)  # remove single bonded atoms
  g.center() # center the geometry
  return g # return the new geometry





def reciprocal(v1,v2,v3=np.array([0.,0.,1.])):
  """Return the reciprocal vectors"""
  vol = v1.dot(np.cross(v2,v3)) # volume
  w1 = np.cross(v2,v3)/vol
  w2 = np.cross(v3,v1)/vol
  w3 = np.cross(v1,v2)/vol
  return (w1,w2,w3)



def build_ribbon(g,n):
  """ Return a geometry of a ribbon based on this cell"""
  if g.dimensionality!=2: raise # if it is not two dimensional
  angle = sculpt.get_angle(g.a1,g.a2)/np.pi*180 # get the angle
  if np.abs(angle-90)<1.: # if it is square
    gout = g.copy() # copy geometry
    gout.dimensionality = 1
    rs = []
    for ir in g.r:
      for i in range(n):
        rs.append(ir+g.a1*i) # append position
    gout.r = rs
    gout.r2xyz() # update
    raise
    return gout


def image2island(impath,g,nsuper=4,size=10,color="black",
        npristine=1,periodic=False):
  """Build an island using a certain image"""
  from PIL import Image
  im = Image.open(impath)
  im = im.convert('RGBA')
  data = np.array(im) # convert to array
  red, green, blue, alpha = data.T # store data for readbility
  if color=="black": #retain the black color
    retain = (red < 20) & (blue < 20) & (green < 20)
  elif color=="red": #retain the black color
    retain = (red > 200) & (blue < 20) & (green < 20)
  elif color=="blue": #retain the black color
    retain = (red < 20) & (blue > 200) & (green < 20)
  elif color=="green": #retain the black color
    retain = (red < 20) & (blue < 20) & (green > 200)
  else: raise # unrecognized
  data[..., :-1][retain.T] = (0, 0, 0) # set as black
  data[..., :-1][np.logical_not(retain.T)] = (255, 255, 255) # set as white
#  data[..., :-1][not retain.T] = (255, 255, 255) # set as black
  im2 = Image.fromarray(data) # convert to image
  im2 = im2.convert("L") # to black and white
  bw = np.asarray(im2).copy() # convert to array
  bw[bw < 128] = 0  # Black
  bw[bw >= 128] = 1 # White
  bw = bw.transpose() # transpose image
  # now create a supercell
  nx,ny = bw.shape # size of the image
  go = g.supercell(nsuper*size) # build supercell
  go.center()
  minx = -size
  maxx = size
  miny = -size*bw.shape[1]/bw.shape[0]
  maxy = size*bw.shape[1]/bw.shape[0]
  def finter(rtmp):
    x = rtmp[0]
    y = rtmp[1]
    x = (x - minx)/(maxx-minx)
    y = (y - miny)/(maxy-miny)
    xi = (nx-1)*x # normalized
    yi = (ny-1)*y # normalized
    xi,yi = int(round(xi)),int(round(yi)) # integer
    if not 0<xi<bw.shape[0]: 
        if periodic: return True
        else: return False
    if not 0<yi<bw.shape[1]:
        if periodic: return True
        else: return False
    if bw[xi,yi]==0: return True
    else: return False
  if periodic: 
      go = go.supercell(npristine) # pristine supercell
  go = intersec(go,finter)
  if not periodic: go.dimensionality = 0 # zero dimensional
  else: go.celldis = None
  return go 


def common(g1,g2,tol=0.1):
  """Return the indexes of atoms common in both structures"""
  indexes = []
  for i in range(len(g1.r)):
    r1 = g1.r[i] # position
    for r2 in g2.r:
      dr = r1-r2
      if dr.dot(dr)<tol:
        indexes.append(i)
        break # next iteration
  return indexes

def add(g1,g2):
  g = g1.copy() # copy geometry
  g.x = np.concatenate([g1.x,g2.x])
  g.y = np.concatenate([g1.y,g2.y])
  g.z = np.concatenate([g1.z,g2.z])
  g.xyz2r()
  g.has_sublattice = False
  return g


def set_xy_plane(g):
  """Modify a geometry so the lattice vectors lie in the xy plane"""
#  if g.dimensionality != 2: raise # only for 2d
  go = g.copy() # copy geometry
  nv = np.cross(g.a1,g.a2)
  nv /= np.sqrt(nv.dot(nv)) # unitary normal vector
  rho = np.sqrt(nv[0]*nv[0] + nv[1]*nv[1]) # planar component
  theta = np.arctan2(rho,nv[2]) # theta angle
  phi = np.arctan2(nv[1],nv[0]) # phi angle
  # matrix that transforms (0,0,1) to that vector
  ct,st = np.cos(theta),np.sin(theta)
  cp,sp = np.cos(phi),np.sin(phi)
  Rt = np.matrix([[ct,0,st],[0.,1.,0],[-st,0,ct]]) # rotate along y
  Rp = np.matrix([[cp,-sp,0.],[sp,cp,0],[0.,0.,1.]]) # rotate along z
  R = Rp*Rt # transforms (0,0,1) to nv
  U = R.I # inverse transformation
  # now transform everything
  def transform(r): return np.array((U*np.matrix(r).T).T)[0] 
  go.a1 = transform(g.a1)
  go.a2 = transform(g.a2)
  go.a3 = transform(g.a3)
  go.r = np.array([transform(ri) for ri in g.r])
  go.r2xyz()
  return go




def sites_in_unit_cell(r,a1,a2,a3,dim=3):
  """Retain position located in the unit cell defined by a1,a2,a3"""
  R = np.array([a1,a2,a3]).T # transformation matrix
  L = algebra.inv(R) # inverse matrix
#  d0 = -np.random.random()*0.001 - .5 # accuracy
  d0 = 0.00234231421 - 0.5 # random number
  d1 = 1.0 + d0 # accuracy
  retain = [] # empty list
  for ri in r: # loop over positions
    rn = L@ri  # transform
    n1,n2,n3 = rn[0],rn[1],rn[2]
    if dim==3:
        if d0<n1<d1 and d0<n2<d1 and d0<n3<d1:
            retain.append(1)
        else: retain.append(0)
    if dim==2:
        if d0<n1<d1 and d0<n2<d1:
            retain.append(1)
        else: retain.append(0)
  retain = np.array(retain)
  return retain

def retain_unit_cell(r,a1,a2,a3,dim=3):
    """Retain position located in the unit cell defined by a1,a2,a3"""
    retain = sites_in_unit_cell(r,a1,a2,a3,dim=3)
    return r[retain==1.]



def indexes_line(g,r0,r1=None,width=2.0):
    """Return the indexes along a certain line"""
    if r1 is None: # workaround
        r1 = np.array(r0)
        r0 = np.array([0.,0.,0.])
    i0 = g.closest_index(np.array(r0)) # first index
    i1 = g.closest_index(np.array(r1)) # first index
    r0 = g.r[i0] # initial point
    r1 = g.r[i1] # final point
    # now rotate the geometry to check which
    # atoms to accept
    gtmp = g.copy() # copy geometry
    gtmp.dimensionality = 0 # this is just a workaround
    gtmp.shift(r0) # shift the geometry
    gtmp = rotate_a2b(gtmp,r1-r0,np.array([1.,0.,0.]))
    # new positions
    r0 = gtmp.r[i0]
    r1 = gtmp.r[i1]
    dr = r1 - r0 # vector between both points
    dy = width # width accepted
    dx = np.sqrt(dr.dot(dr))+0.1
    # and every coordinate
    x1,y1 = r0[0],r0[1]
    x2,y2 = r1[0],r1[1]
    ym = (y1+y2)/2. # average y
    def fun(r):
      """Function that decides which atoms to calculate"""
      x0 = r[0]
      y0 = r[1]
      if (x1-dy)<x0<(x2+dy) and np.abs(y0-ym)<dy: return True
      else: return False
    inds = intersected_indexes(gtmp,fun) # indexes of the atoms
    ur = dr/np.sqrt(dr.dot(dr)) # unitary vector
    steps = [(gtmp.r[i] - r0).dot(ur) for i in inds] # proyect along that line
    inds = [i for (s,i) in sorted(zip(steps,inds))] # sort by step
    steps = sorted(steps) # sort the steps
    return inds,np.array(steps) # return the index and associated steps








